# Copyright (C) 2021-2022 Xilinx, Inc.  All rights reserved.
# Copyright (C) 2022-2024 Advanced Micro Devices, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-3-Clause

__author__ = "Anish Kadamathikuttiyil Karthikeyan Pillai"
__copyright__ = "Copyright 2021, Xilinx"

import sys
def usage():
    # sys.argv[0] - usage python file name
    # sys.argv[1] - xclient/xcffi/xpyro
    # sys.argv[2] - host/board
    # sys.argv[3] - ip_address
    # sys.argv[4] - port
    # sys.argv[5] - "switchable" - switchable option integrated in xsa
    # sys.argv[6] - "auxiliary" - auxiliary option integrated in xsa
    print(f"Usage: The application work in three modes xclient, xcffi and xpyro\n"
          f"python3 {sys.argv[0]} xclient host/board ip_address port switchable auxiliary => Runs at host or board with xclient\n"
          f"sudo python3 {sys.argv[0]} xcffi switchable auxiliary => Runs at board with xcffi\n"
          f"python3 {sys.argv[0]} xpyro host/board ip_address port switchable auxiliary => Runs at host or board with xpyro\n"
          f"Example1: python3 {sys.argv[0]} xclient host 169.254.10.2 9090 \n"
          f"Example2: python3 {sys.argv[0]} xclient host 169.254.10.2 9090 no auxiliary\n"
          f"Example3: python3 {sys.argv[0]} xclient host 169.254.10.2 9090 switchable auxiliary")

switchable_flag = False
auxiliary_flag = False

# 'xcffi' option will run only in the board
if (len(sys.argv) ==  2):
    if (sys.argv[1] != 'xcffi'):
        usage()
        sys.exit()
elif (len(sys.argv) ==  4):
    if (sys.argv[1] != 'xcffi'):
        usage()
        sys.exit()
    else:
        if (sys.argv[2] == 'switchable'):
            switchable_flag = True
        if (sys.argv[3] == 'auxiliary'):
            auxiliary_flag = True
elif (len(sys.argv) ==  5):
    pass
elif (len(sys.argv) ==  7):
    if (sys.argv[5] == 'switchable'):
        switchable_flag = True
    if (sys.argv[6] == 'auxiliary'):
        auxiliary_flag = True
else:
    usage()
    sys.exit()

# The 'xclient' option will run both in host and board
if (sys.argv[1] == 'xclient'):
    # For 'xclient' option the path of the xclient is required
    # The relative path in the host is given for host case
    if (sys.argv[2] == 'host'):
        sys.path.append('../../../xclient/dfe')
    # The location of xclient in the filesystem is given for the board case
    elif (sys.argv[2] == 'board'):
        sys.path.append('/usr/share/raft/xclient/dfe')
    else:
        usage()
        sys.exit()
    import mix_client
    handle = mix_client.mix
    ip_address = sys.argv[3]
    port = sys.argv[4]

elif (sys.argv[1] == 'xcffi'):
    # For 'xcffi' option the path of the cffi python code is required
    sys.path.append('/usr/share/raft/xserver/xcffi/drv_api/dfe')
    from mix_server import MIX
    handle = MIX()

# The 'xpyro' option will run both in host and board
elif (sys.argv[1] == 'xpyro'):
    ip_address = sys.argv[3]
    port = sys.argv[4]
    import Pyro4
    # Prepare the uri needed for pyro communication
    uri = f"PYRO:MIX@{ip_address}:{port}"
    handle = Pyro4.Proxy(uri)
else:
    usage()
    sys.exit()

if (sys.argv[1] == 'xclient'):
    #SetIpAndPort
    handle.SetIpAndPort(ip_address, port)
    #GetPythonLogLevels
    PythonLogLevels = handle.GetPythonLogLevels()
    #SetClientLogLevel
    handle.SetClientLogLevel(PythonLogLevels["DEBUG"])

#Description:
#   Return the logging levels supported by logging library in python
#Input Arguments:
#   None
#Return:
#   Dictionary showing the log levels supported by logging library

#GetPythonLogLevels
PythonLogLevels = handle.GetPythonLogLevels()

#Description:
#   Set the python log level to the given level
#Input Arguments:
#   Log level to set
#Return:
#   None

#SetServerLogLevel
handle.SetServerLogLevel(PythonLogLevels["DEBUG"])

#Description:
#   Set the metal log level to the given level
#Input Arguments:
#   Log level to set
#Return:
#   None

#SetMetalLogLevel
metal_log_level = handle.GetEnum_metal_log_level()
handle.SetMetalLogLevel(metal_log_level["METAL_LOG_EMERGENCY"])

#Description:
#   Return Dictionary with all MIX macros in the mix header file
#Input Arguments:
#   None
#Return:
#   Dictionary with all MIX macros in the mix header file
#GetMixMacro
mix_macro = handle.GetMixMacro()
print(mix_macro["XDFEMIX_MAX_NUM_INSTANCES"])
print(mix_macro["XST_SUCCESS"])
print(mix_macro["XST_FAILURE"])
print(mix_macro["XDFEMIX_NODE_NAME_MAX_LENGTH"])
print(mix_macro["XDFEMIX_CC_NUM"])
print(mix_macro["XDFEMIX_ANT_NUM_MAX"])
print(mix_macro["XDFEMIX_SEQ_LENGTH_MAX"])
print(mix_macro["XDFEMIX_RATE_MAX"])
print(mix_macro["XDFEMIX_AUX_NCO_MAX"])
print(mix_macro["XDFEMIX_NCO_MAX"])
print(mix_macro["XDFEMIX_CC_GAIN_MAX"])

#Description:
#   Return Dictionary equivalent of enum XDfeMix_StateId
#Input Arguments:
#   None
#Return:
#   Dictionary equivalent of enum XDfeMix_StateId

#GetEnum_XDfeMix_StateId
XDfeMix_StateId = handle.GetEnum_XDfeMix_StateId()

#Description:
#   Return Dictionary equivalent of enum metal_log_level
#Input Arguments:
#   None
#Return:
#   Dictionary equivalent of enum metal_log_level

#GetEnum_metal_log_level
metal_log_level = handle.GetEnum_metal_log_level()

#Description:
#   API initialises one instance of a Mixer driver.
#   Traverses "/sys/bus/platform/device" directory (in Linux), to find registered
#   XDfeMix device with the name DeviceNodeName. The first available slot in
#   the instances array XDfeMix_Mixer[] will be taken as a DeviceNodeName
#   object. On success it moves the state machine to a Ready state, while on
#   failure stays in a Not Ready state.
#C header declaration:
#   XDfeMix *XDfeMix_InstanceInit(const char *DeviceNodeName);
#Input Arguments:
#   DeviceNodeName: device node name.
#Return:
#   ret - 0 on success, 1 on failure
#   device_id - integer handle to initialized instance
#   DeviceNodeNameRet - device node name returned from driver
#                       which will be same as the passed value

DeviceNodeName = "a7c40000.xdfe_cc_mixer"
ret, device_id, DeviceNodeName = handle.XDfeMix_InstanceInit(DeviceNodeName)

#Description:
#   Writes a value to register in a Mixer instance.
#C header declaration:
#   void XDfeMix_WriteReg(const XDfeMix *InstancePtr, u32 AddrOffset, u32 Data);
#Input Arguments:
#   device_id: id of the opened device.
#   AddrOffset: is address offset relative to instance base address.
#   Data: is value to be written.
#Return:
#   None

handle.XDfeMix_WriteReg(device_id, 0x20, 0x1001)

#Description:
#   Reads a value from the register from a Mixer instance.
#C header declaration:
#   u32 XDfeMix_ReadReg(const XDfeMix *InstancePtr, u32 AddrOffset);
#Input Arguments:
#   device_id: id of the opened device.
#   AddrOffset: is address offset relative to instance base address.
#Return:
#   regval: Register value.

ret = handle.XDfeMix_ReadReg(device_id, 0)

#Description:
#   Resets and put block into a reset state.
#C header declaration:
#   void XDfeMix_Reset(XDfeMix *InstancePtr);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   None

handle.XDfeMix_Reset(device_id)
handle.XDfeMix_WriteReg(device_id, 0x20, 0x1001)

#Description:
#   Reads configuration from device tree/xparameters.h and IP registers.
#   Removes S/W reset and moves the state machine to a Configured state.
#C header declaration:
#   void XDfeMix_Configure(XDfeMix *InstancePtr, XDfeMix_Cfg *Cfg);
#Input Arguments:
#   device_id: id of the opened device.
#   Cfg: configuration data container.
#Return:
#   Cfg: configuration data container.

Cfg_in = handle.GetStruct_XDfeMix_Cfg()
Cfg = handle.XDfeMix_Configure(device_id, Cfg_in)
handle.XDfeMix_WriteReg(device_id, 0x20, 0x1001)

#Description:
#   DFE Mixer driver one time initialisation and moves the state machine to
#   an Initialised state.
#C header declaration:
#   void XDfeMix_Initialize(XDfeMix *InstancePtr, const XDfeMix_Init *Init);
#Input Arguments:
#   device_id: id of the opened device.
#   Init: initialisation data container.
#Return:
#   Init: initialisation data container.

Init_in = handle.GetStruct_XDfeMix_Init()
Init_in['Sequence']['Length']=6
Init_out = handle.XDfeMix_Initialize(device_id, Init_in)

#Description:
#   Sets trigger configuration.
#C header declaration:
#   void XDfeMix_SetTriggersCfg(const XDfeMix *InstancePtr, XDfeMix_TriggerCfg *TriggerCfg);
#Input Arguments:
#   device_id: id of the opened device.
#   TriggerCfg: trigger configuration container.
#Return:
#   TriggerCfg: trigger configuration container.

TriggerCfg_in = handle.GetStruct_XDfeMix_TriggerCfg()
handle.XDfeMix_SetTriggersCfg(device_id, TriggerCfg_in)

#Description:
#   Activates Mixer and moves the state machine to an Activated state.
#C header declaration:
#   void XDfeMix_Activate(XDfeMix *InstancePtr, bool EnableLowPower);
#Input Arguments:
#   device_id: id of the opened device.
#   EnableLowPower: flag indicating low power.
#Return:
#   None

handle.XDfeMix_Activate(device_id, 0)

#Description:
#   Gets a state machine state id. The state id is returned as a string.
#   The enum is mapped to a dictionary in python.
#C header declaration:
#   XDfeMix_StateId XDfeMix_GetStateID(XDfeMix *InstancePtr);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   StateId: State machine StateID as a string

StateId = handle.XDfeMix_GetStateID(device_id)
print(StateId)

#Description:
#   Returns the current CC configuration. Not used slot ID in a sequence
#   (Sequence.CCID[Index]) are represented as (-1), not the value in registers.
#C header declaration:
#   void XDfeMix_GetCurrentCCCfg(const XDfeMix *InstancePtr, XDfeMix_CCCfg *CCCfg);
#Input Arguments:
#   device_id: id of the opened device.
#   CurrCCCfg: CC configuration container.
#Return:
#   CurrCCCfg: CC configuration container

CurrCCCfg = handle.GetStruct_XDfeMix_CCCfg()
CurrentCCCfg = handle.XDfeMix_GetCurrentCCCfg(device_id, CurrCCCfg)

#Description:
#   Returns the current CC configuration for DL and UL in switchable mode.
#   Not used slot ID in a sequence (Sequence.CCID[Index]) are represented
#   as (-1), not the value in registers.
#C header declaration:
#   void XDfeMix_GetCurrentCCCfgSwitchable(const XDfeMix *InstancePtr,
#                 XDfeMix_CCCfg *CCCfgDownlink,
#                 XDfeMix_CCCfg *CCCfgUplink);
#Input Arguments:
#   device_id: id of the opened device.
#   CCCfgDownlink: Downlink CC configuration container.
#   CCCfgUplink: Uplink CC configuration container.
#Return:
#   CCCfgDownlink, CCCfgUplink: CC configuration container

if (switchable_flag == True):
    CCCfgDownlink = handle.GetStruct_XDfeMix_CCCfg()
    CCCfgUplink = handle.GetStruct_XDfeMix_CCCfg()
    CCCfgDownlink, CCCfgUplink = handle.XDfeMix_GetCurrentCCCfgSwitchable(device_id, CCCfgDownlink, CCCfgUplink)

#Description:
#   Returns configuration structure CCCfg with CCCfg->Sequence.Length value set
#   in XDfeMix_Configure(), array CCCfg->Sequence.CCID[] members are set to not
#   used value (-1) and the other CCCfg members are set to 0.
#C header declaration:
#   void XDfeMix_GetEmptyCCCfg(const XDfeMix *InstancePtr, XDfeMix_CCCfg *CCCfg);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   CCCfg: CC configuration container

CCCfg = handle.XDfeMix_GetEmptyCCCfg(device_id)

#Description:
#   Return Dictionary equivalent of structure XDfeMix_InternalDUCDDCCfg
#Input Arguments:
#   None
#Return:
#   Dictionary equivalent of structure XDfeMix_InternalDUCDDCCfg

#GetStruct_XDfeMix_InternalDUCDDCCfg
XDfeMix_InternalDUCDDCCfg = handle.GetStruct_XDfeMix_InternalDUCDDCCfg()

#Description:
#   Returns the current CC sequence bitmap, CCID carrier configuration and
#   NCO configuration.
#C header declaration:
#   void XDfeMix_GetCarrierCfgAndNCO(const XDfeMix *InstancePtr, XDfeMix_CCCfg *CCCfg,
#   s32 CCID, u32 *CCSeqBitmap, XDfeMix_CarrierCfg *CarrierCfg, XDfeMix_NCO *NCO);
#Input Arguments:
#   device_id: id of the opened device.
#   CCCfg: component carrier (CC) configuration container.
#   CCID: Channel ID.
#Return:
#   CCSeqBitmap: CC slot position container.
#   CarrierCfg: CC configuration container.
#   NCO: NCO configuration container.

CCCfg = handle.GetStruct_XDfeMix_CCCfg()
CCSeqBitmap, CarrierCfg, NCO = handle.XDfeMix_GetCarrierCfgAndNCO(device_id, CCCfg, 0)

#Description:
#   Set antenna configuration in CC configuration container.
#C header declaration:
#   void XDfeMix_SetAntennaCfgInCCCfg(const XDfeMix *InstancePtr,
#    XDfeMix_CCCfg *CCCfg, u32 *AntennaCfg);
#Input Arguments:
#   device_id: id of the opened device.
#   CCCfg: component carrier (CC) configuration container.
#   AntennaCfg: Array of all antenna configurations.
#Return:
#   CCCfg: component carrier (CC) configuration container.

AntennaCfg = handle.GetStruct_XDfeMix_AntennaCfg()
AntennaCfg = {
    "Gain": [1,1,1,1,1,1,0,0],
}
CCCfg = handle.XDfeMix_SetAntennaCfgInCCCfg(device_id, CCCfg, AntennaCfg)

#Description:
#   Updates antenna configuration to all antennas.
#C header declaration:
#   u32 XDfeMix_UpdateAntennaCfg(XDfeMix *InstancePtr,
#   XDfeMix_AntennaCfg *AntennaCfg);
#Input Arguments:
#   device_id: id of the opened device.
#   AntennaCfg: Array of all antenna configurations.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#Note:
#   Clear event status with XDfeMix_ClearEventStatus() before running this API.

AntennaCfg = handle.GetStruct_XDfeMix_AntennaCfg()
AntennaCfg = {
    "Gain": [1,1,1,1,1,1,0,0],
}
ret = handle.XDfeMix_UpdateAntennaCfg(device_id, AntennaCfg)

#Description:
#   Adds specified CCID, with specified configuration, to a local CC
#   configuration structure.
#   If there is insufficient capacity for the new CC the function will return
#   an error.
#   Initiates CC update (enable CCUpdate trigger TUSER Single Shot).
#
#   The returned CCCfg.Sequence is transleted as there is no explicit indication
#   that SEQUENCE[i] is not used - 0 can define the slot as either used or
#   not used. Sequence data that is returned in the CCIDSequence is not the same
#   as what is written in the registers. The translation is:
#   - CCIDSequence.CCID[i] = -1    - if [i] is unused slot
#   - CCIDSequence.CCID[i] = CCID  - if [i] is used slot
#   - a returned CCIDSequence->Length = length in register + 1
#C header declaration:
#   u32 XDfeMix_AddCCtoCCCfg(XDfeMix *InstancePtr, XDfeMix_CCCfg *CCCfg, s32 CCID,
#   u32 CCSeqBitmap, const XDfeMix_CarrierCfg *CarrierCfg);
#Input Arguments:
#   device_id: id of the opened device.
#   CCCfg: component carrier (CC) configuration container.
#   CCID: Channel ID.
#   CCSeqBitmap: CC slot position container.
#   CarrierCfg: CC configuration container.
#   NCO: NCO configuration container.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#   CCCfg: component carrier (CC) configuration container.

CarrierCfg = handle.GetStruct_XDfeMix_CarrierCfg()
NCO = handle.GetStruct_XDfeMix_NCO()
CCID = 0x0
CCSeqBitmap = 0x1
CurrCCCfg = handle.GetStruct_XDfeMix_CCCfg()
CurrentCCCfg = handle.XDfeMix_GetCurrentCCCfg(int(device_id), CurrCCCfg)
ret, CurrentCCCfg = handle.XDfeMix_AddCCtoCCCfg(int(device_id), CurrentCCCfg, CCID, CCSeqBitmap, CarrierCfg, NCO)

#Description:
#   Removes specified CCID from a local CC configuration structure.
#C header declaration:
#   void XDfeMix_RemoveCCfromCCCfg(XDfeMix *InstancePtr, XDfeMix_CCCfg *CCCfg,
#   s32 CCID);
#Input Arguments:
#   device_id: id of the opened device.
#   CCCfg: component carrier (CC) configuration container.
#   CCID: Channel ID.
#Return:
#   CCCfg: component carrier (CC) configuration container.

CCCfg = handle.GetStruct_XDfeMix_CCCfg()
CCCfg = handle.XDfeMix_RemoveCCfromCCCfg(device_id, CCCfg, 0)

#Description:
#   Updates specified CCID, with specified configuration to a local CC
#   configuration structure.
#   If there is insufficient capacity for the new CC the function will return
#   an error.
#C header declaration:
#   void XDfeMix_UpdateCCinCCCfg(const XDfeMix *InstancePtr, XDfeMix_CCCfg *CCCfg,
#   s32 CCID, const XDfeMix_CarrierCfg *CarrierCfg);
#Input Arguments:
#   device_id: id of the opened device.
#   CCCfg: component carrier (CC) configuration container.
#   CCID: Channel ID.
#   CarrierCfg: CC configuration container.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#   CarrierCfg: is a CC configuration container.

CCCfg = handle.GetStruct_XDfeMix_CCCfg()
CarrierCfg = handle.GetStruct_XDfeMix_CarrierCfg()
ret, CCCfg = handle.XDfeMix_UpdateCCinCCCfg(device_id, CCCfg, 0, CarrierCfg)

#Description:
#   Writes local CC configuration to the shadow (NEXT) registers and triggers
#   copying from shadow to operational registers.
#C header declaration:
#   u32 XDfeMix_SetNextCCCfgAndTrigger(const XDfeMix *InstancePtr,
#    XDfeMix_CCCfg *CCCfg);
#Input Arguments:
#   device_id: id of the opened device.
#   CCCfg: component carrier (CC) configuration container.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#   CCCfg: component carrier (CC) configuration container.

CCCfg = handle.GetStruct_XDfeMix_CCCfg()
ret, CCCfg = handle.XDfeMix_SetNextCCCfgAndTrigger(device_id, CCCfg)

#Description:
#   Writes local CC configuration to the shadow (NEXT) registers and triggers
#   copying from shadow to operational (CURRENT) registers.
#C header declaration:
#   u32 XDfeMix_SetNextCCCfgAndTriggerSwitchable(const XDfeMix *InstancePtr,
#                         XDfeMix_CCCfg *CCCfgDownlink,
#                         XDfeMix_CCCfg *CCCfgUplink)
#Input Arguments:
#   device_id: id of the opened device.
#   CCCfgDownlink: Downlink CC configuration container.
#   CCCfgUplink: Uplink CC configuration container.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#   CCCfgDownlink: Downlink CC configuration container.
#   CCCfgUplink: Uplink CC configuration container.

if (switchable_flag == True):
    CCCfgDownlink = handle.GetStruct_XDfeMix_CCCfg()
    CCCfgUplink = handle.GetStruct_XDfeMix_CCCfg()
    ret, CCCfgDownlink, CCCfgUplink = handle.XDfeMix_SetNextCCCfgAndTriggerSwitchable(device_id, CCCfgDownlink, CCCfgUplink)

#Description:
#   Adds specified CCID, with specified configuration.
#   If there is insufficient capacity for the new CC the function will return
#   an error.
#   Initiates CC update (enable CCUpdate trigger TUSER Single Shot).
#C header declaration:
#   u32 XDfeMix_AddCC(XDfeMix *InstancePtr, s32 CCID, u32 CCSeqBitmap,
#    const XDfeMix_CarrierCfg *CarrierCfg, const XDfeMix_NCO *NCO);
#Input Arguments:
#   device_id: id of the opened device.
#   CCID: is a Channel ID.
#   BitSequence: up to 16 defined slots into which a CC can be
#                allocated. The number of slots can be from 1 to 16 depending on
#                system initialization. The number of slots is defined by the
#                "sequence length" parameter which is provided during initialization.
#                The Bit offset within the CCSeqBitmap indicates the equivalent
#                Slot number to allocate. e.g. 0x0003  means the caller wants the
#                passed component carrier (CC) to be allocated to slots 0 and 1.
#   CarrierCfg: is a CC configuration container.
#   NCO: NCO configuration container.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#   CarrierCfg: is a CC configuration container.
#Note:
#   Clear event status with XDfeMix_ClearEventStatus() before running this API.

CarrierCfg_in = handle.GetStruct_XDfeMix_CarrierCfg()
BitSequence = 0xb
NCO = handle.GetStruct_XDfeMix_NCO()
ret, CarrierCfg_out = handle.XDfeMix_AddCC(device_id, 0, BitSequence, CarrierCfg_in, NCO)

#Description:
#   Removes specified CCID.
#   Initiates CC update (enable CCUpdate trigger TUSER Single Shot).
#C header declaration:
#   u32 XDfeMix_RemoveCC(const XDfeMix *InstancePtr, u32 CCID);
#Input Arguments:
#   device_id: id of the opened device.
#   CCID: is a Channel ID.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#Note:
#   Clear event status with XDfeMix_ClearEventStatus() before running this API.

ret = handle.XDfeMix_RemoveCC(device_id, 0)

#Description:
#   Moves specified CCID from one NCO to another aligning phase to make it
#   transparent.
#   Initiates CC update (enable CCUpdate trigger TUSER Single Shot).
#C header declaration:
#   u32 XDfeMix_MoveCC(const XDfeMix *InstancePtr, u32 CCID, u32 Rate,
#   u32 FromNCO, u32 ToNCO);
#Input Arguments:
#   device_id: id of the opened device.
#   CCID: is a Channel ID.
#   Rate: is a NCO rate value.
#   FromNCO: is a NCO value moving from.
#   ToNCO: is a NCO value moving to.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#Note:
#   Clear event status with XDfeMix_ClearEventStatus() before running this API.

ret = handle.XDfeMix_MoveCC(device_id, 0, 0, 0, 0)

#Description:
#   Updates specified CCID, with a configuration defined in CarrierCfg
#   structure.
#   If there is insufficient capacity for the new CC the function will return
#   an error.
#C header declaration:
#   u32 XDfeMix_UpdateCC(const XDfeMix *InstancePtr, s32 CCID,
#    const XDfeMix_CarrierCfg *CarrierCfg);
#Input Arguments:
#   device_id: id of the opened device.
#   CCID: is a Channel ID.
#   CarrierCfg: is a CC configuration container.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.

CarrierCfg_in = handle.GetStruct_XDfeMix_CarrierCfg()
ret = handle.XDfeMix_UpdateCC(device_id, 0, CarrierCfg_in)

#Description:
#   Sets antenna gain. Initiate CC update (enable CCUpdate trigger TUSER
#   Single Shot).
#C header declaration:
#   u32 XDfeMix_SetAntennaGain(const XDfeMix *InstancePtr, u32 AntennaId, u32 AntennaGain);
#Input Arguments:
#   device_id: id of the opened device.
#   AntennaId: is an antenna ID.
#   AntennaGain: is an antenna gain.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#Note:
#   Clear event status with XDfeMix_ClearEventStatus() before running this API.

ret = handle.XDfeMix_SetAntennaGain(device_id, 0, 0)

#Description:
#   Returns current trigger configuration.
#C header declaration:
#   void XDfeMix_GetTriggersCfg(const XDfeMix *InstancePtr, XDfeMix_TriggerCfg *TriggerCfg);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   TriggerCfg: trigger configuration container

TriggerCfg_out = handle.XDfeMix_GetTriggersCfg(device_id)

#Description:
#   Gets DUC/DDC overflow status.
#C header declaration:
#   void XDfeMix_GetDUCDDCStatus(const XDfeMix *InstancePtr, XDfeMix_DUCDDCStatus *DUCDDCStatus);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   DUCDDCStatus: DUC/DDC status container.

DUCDDCStatus_out = handle.XDfeMix_GetDUCDDCStatus(device_id)

#Description:
#   Gets Mixer status for a specified CCID.
#C header declaration:
#   void XDfeMix_GetMixerStatus(const XDfeMix *InstancePtr, XDfeMix_MixerStatus *MixerStatus);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   DUCDDCStatus: DUC/DDC status container.

MixerStatus_out = handle.XDfeMix_GetMixerStatus(device_id)

#Description:
#   Gets interrupt mask status.
#C header declaration:
#   void XDfeMix_GetInterruptMask(const XDfeMix *InstancePtr,
#   XDfeMix_InterruptMask *Mask);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   Mask: interrupt masks container.

InterruptMask_out = handle.XDfeMix_GetInterruptMask(device_id)

#Description:
#   Sets interrupt mask.
#C header declaration:
#   void XDfeMix_SetInterruptMask(const XDfeMix *InstancePtr,
#   const XDfeMix_InterruptMask *Mask);
#Input Arguments:
#   device_id: id of the opened device.
#   Mask: interrupt masks flags container.
#         0 - does not mask corresponding interrupt
#         1 - masks corresponding interrupt
#Return:
#   None

InterruptMask_in = handle.GetStruct_XDfeMix_InterruptMask()
InterruptMask_in['DUCDDCOverflow'] = 1
InterruptMask_in['MixerOverflow'] = 1
InterruptMask_in['CCUpdate'] = 1
InterruptMask_in['CCSequenceError'] = 1
handle.XDfeMix_SetInterruptMask(device_id, InterruptMask_in)

#Description:
#   Sets uplink/downlink register bank.
#C header declaration:
#   void XDfeMix_SetRegBank(const XDfeMix *InstancePtr,
#                  u32 RegBank)
#Input Arguments:
#   device_id: id of the opened device.
#   RegBank: Register bank value to be set.
#Return:
#   None

if (switchable_flag == True):
    RegBank = 1
    handle.XDfeMix_SetRegBank(device_id, RegBank)
    RegBank = 0
    handle.XDfeMix_SetRegBank(device_id, RegBank)

#Description:
#   Gets event status.
#C header declaration:
#   void XDfeMix_GetEventStatus(const XDfeMix *InstancePtr, XDfeMix_Status *Status);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   Status: event status container.

EventStatus_out = handle.XDfeMix_GetEventStatus(device_id)

#Description:
#   Clears event status.
#C header declaration:
#   void XDfeMix_ClearEventStatus(const XDfeMix *InstancePtr, const XDfeMix_Status *Status);
#Input Arguments:
#   device_id: id of the opened device.
#   Status: Clear event status container.
#           0 - does not clear corresponding event status
#           1 - clear corresponding event status
#Return:
#   None

Status_in = handle.GetStruct_XDfeMix_Status()
Status_in['DUCDDCOverflow'] = 1
Status_in['MixerOverflow'] = 1
Status_in['CCUpdate'] = 1
Status_in['CCSequenceError'] = 1
handle.XDfeMix_ClearEventStatus(device_id, Status_in)

#Description:
#   Reads the delay, which will be added to TUSER and TLAST (delay matched
#   through the IP).
#C header declaration:
#   u32 XDfeMix_GetTUserDelay(const XDfeMix *InstancePtr);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   ret: Delay value

ret = handle.XDfeMix_GetTUserDelay(device_id)

#Description:
#   Returns data latency + tap.
#C header declaration:
#   u32 XDfeMix_GetTDataDelay(const XDfeMix *InstancePtr, u32 Tap, u32 *TDataDelay);
#Input Arguments:
#   device_id: id of the opened device.
#   Tap: Tap value.
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#   TDataDelay: Returned Data latency value.

Tap = 0
ret, TDataDelay = handle.XDfeMix_GetTDataDelay(device_id, Tap)

#Description:
#   Returns predefined Central Tap value for chosen RATE.
#C header declaration:
#   u32 XDfeMix_GetCenterTap(const XDfeMix *InstancePtr, u32 Rate, u32 *CenterTap);
#Input Arguments:
#   device_id: id of the opened device.
#   Rate: Interpolation/decimation rate index value [1-5].
#Return:
#   ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
#   CenterTap: Returned Central Tap value.

Rate = 1
ret, CenterTap = handle.XDfeMix_GetCenterTap(device_id, Rate)

#Description:
#   Deactivates Mixer and moves the state machine to Initialised state.
#C header declaration:
#   void XDfeMix_GetVersions(const XDfeMix *InstancePtr, XDfeMix_Version *SwVersion,
#    XDfeMix_Version *HwVersion);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   None

SwVersion_out, HwVersion_out = handle.XDfeMix_GetVersions(device_id)

#Description:
#   Adds specified auxiliary NCO, with specified configuration, to a local CCCfg.
#C header declaration:
#   void XDfeMix_AddAuxNCOtoCCCfg(XDfeMix *InstancePtr, XDfeMix_CCCfg *CCCfg,
#                  const s32 AuxId, const XDfeMix_NCO *NCO,
#                  const XDfeMix_AuxiliaryCfg *AuxCfg)
#Input Arguments:
#   device_id: id of the opened device.
#   CCCfg: component carrier (CC) configuration container.
#   AuxId: Channel ID.
#   NCO: Auxiliary NCO configuration container.
#   AuxCfg: Auxiliary NCO configuration container.
#Return:
#   CCCfg: component carrier (CC) configuration container.
if (auxiliary_flag == True):
    CurrCCCfg = handle.GetStruct_XDfeMix_CCCfg()
    CurrentCCCfg = handle.XDfeMix_GetCurrentCCCfg(int(device_id), CurrCCCfg)
    NCO = handle.GetStruct_XDfeMix_NCO()
    AuxiliaryCfg = handle.GetStruct_XDfeMix_AuxiliaryCfg()
    AuxId = 1
    CurrentCCCfg = handle.XDfeMix_AddAuxNCOtoCCCfg(device_id, CurrentCCCfg, AuxId, NCO, AuxiliaryCfg)
    AuxId = 2
    CurrentCCCfg = handle.XDfeMix_AddAuxNCOtoCCCfg(device_id, CurrentCCCfg, AuxId, NCO, AuxiliaryCfg)

#Description:
#   Disables specified auxiliary NCO configuration structure.
#C header declaration:
#   void XDfeMix_RemoveAuxNCOfromCCCfg(XDfeMix *InstancePtr, XDfeMix_CCCfg *CCCfg,
#                  const s32 AuxId)
#Input Arguments:
#   device_id: id of the opened device.
#   CCCfg: component carrier (CC) configuration container.
#   AuxId: Channel ID.
#Return:
#   CCCfg: component carrier (CC) configuration container.
if (auxiliary_flag == True):
    AuxId = 1
    CurrentCCCfg = handle.XDfeMix_RemoveAuxNCOfromCCCfg(device_id, CurrentCCCfg, AuxId)

#Description:
#   Deactivates Mixer and moves the state machine to Initialised state.
#C header declaration:
#   void XDfeMix_Deactivate(XDfeMix *InstancePtr);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   None

handle.XDfeMix_Deactivate(device_id)

#Description:
#   Sets the delay, which will be added to TUSER and TLAST (delay matched
#   through the IP).
#C header declaration:
#   u32 XDfeMix_SetTUserDelay(const XDfeMix *InstancePtr, u32 Delay);
#Input Arguments:
#   device_id: id of the opened device.
#   Delay: requested delay variable.
#Return:
#   None

handle.XDfeMix_SetTUserDelay(device_id, 10)

#Description:
#   API closes the instances of a Mixer driver and moves the state machine to
#   a Not Ready state.
#C header declaration:
#   void XDfeMix_InstanceClose(XDfeMix *InstancePtr);
#Input Arguments:
#   device_id: id of the opened device.
#Return:
#   None

handle.XDfeMix_InstanceClose(device_id)



# Copyright (C) 2021-2022 Xilinx, Inc.  All rights reserved.
# Copyright (C) 2022-2024 Advanced Micro Devices, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-3-Clause

__author__ = "Anish Kadamathikuttiyil Karthikeyan Pillai"
__copyright__ = "Copyright 2021, Xilinx"

import sys
RAFT_DIR = '/usr/share/raft/'
sys.path.append(RAFT_DIR + 'xserver/utils')
import logging
import json
from utils import ffi, open_c_library, cdata_to_py
from utils import extract_b64_encoded_string, cdata_string_to_py, xhelper_handle
from utils import get_python_log_levels, getkey_from_listbeginvalue

equ_handle = open_c_library(RAFT_DIR + "xserver/xcffi/drv_header/dfe/xdfeequ_python.h", "/usr/lib/libdfeequ.so.1")


class EQU(object):
    equ_dict = {}
    device_id = 0
    logger = None

    def __init__(self):
        self.logger = self.GetLogger()
        ret = xhelper_handle.XHelper_MetalInit(xhelper_handle.METAL_LOG_ERROR)
        if 0 != ret:
            self.logger.error(f"EQU: XHelper_MetalInit failed. ret = {ret}")
        self.logger.info("Inside EQU Constructor")
        pass

    @staticmethod
    def GetLogger():
        """
        Static method to get the logger for the class.
        Default loglevel is set inside this class

        :return: logger

        """
        log_level = logging.ERROR
        logging.basicConfig(format="%(levelname)s:%(message)s")
        logger = logging.getLogger(__name__)
        try:
            handler_set_check = getattr(logger, 'handler_set')
        except AttributeError:
            handler_set_check = False
        if not handler_set_check:
            logger.setLevel(log_level)
            logger.handler_set = True
            logger.disabled = False
        return logger

    # Log level
    def GetPythonLogLevels(self):
        """
        Return the logging levels supported by logging library in python

        :param : None
        :return: Dictionary showing the log levels supported by logging library
        """
        return get_python_log_levels()

    def SetServerLogLevel(self, PythonLogLevel):
        """
        Set the python log level to the given level

        :param : Log level to set
        :return: None
        """
        self.logger.debug(f"PythonLogLevel = {PythonLogLevel}")
        LogLevelsDict = get_python_log_levels()
        if PythonLogLevel == LogLevelsDict["DEBUG"]:
            self.logger.setLevel(logging.DEBUG)
        elif PythonLogLevel == LogLevelsDict["INFO"]:
            self.logger.setLevel(logging.INFO)
        elif PythonLogLevel == LogLevelsDict["WARNING"]:
            self.logger.setLevel(logging.WARNING)
        elif PythonLogLevel == LogLevelsDict["ERROR"]:
            self.logger.setLevel(logging.ERROR)
        else:
            self.logger.setLevel(logging.CRITICAL)
        return

    def SetMetalLogLevel(self, MetalLogLevel):
        """
        Set the metal log level to the given level

        :param : Log level to set
        :return: None
        """
        self.logger.debug(f"SetMetalLogLevel({MetalLogLevel})")
        metal_log_level = ffi.typeof("enum metal_log_level").relements
        if MetalLogLevel == metal_log_level["METAL_LOG_DEBUG"]:
            xhelper_handle.XHelper_MetalSetLogLevel(xhelper_handle.METAL_LOG_DEBUG)
        elif MetalLogLevel == metal_log_level["METAL_LOG_INFO"]:
            xhelper_handle.XHelper_MetalSetLogLevel(xhelper_handle.METAL_LOG_INFO)
        elif MetalLogLevel == metal_log_level["METAL_LOG_NOTICE"]:
            xhelper_handle.XHelper_MetalSetLogLevel(xhelper_handle.METAL_LOG_NOTICE)
        elif MetalLogLevel == metal_log_level["METAL_LOG_WARNING"]:
            xhelper_handle.XHelper_MetalSetLogLevel(xhelper_handle.METAL_LOG_WARNING)
        elif MetalLogLevel == metal_log_level["METAL_LOG_ERROR"]:
            xhelper_handle.XHelper_MetalSetLogLevel(xhelper_handle.METAL_LOG_ERROR)
        elif MetalLogLevel == metal_log_level["METAL_LOG_CRITICAL"]:
            xhelper_handle.XHelper_MetalSetLogLevel(xhelper_handle.METAL_LOG_CRITICAL)
        elif MetalLogLevel == metal_log_level["METAL_LOG_ALERT"]:
            xhelper_handle.XHelper_MetalSetLogLevel(xhelper_handle.METAL_LOG_ALERT)
        else:
            xhelper_handle.XHelper_MetalSetLogLevel(xhelper_handle.METAL_LOG_EMERGENCY)
        return

    def GetEquMacro(self):
        """
        Return Dictionary with all EQU macros in the equ header file

        :param : None
        :return: Dictionary with all EQU macros in the equ header file
        """
        equ_macro = {}
        equ_macro["XDFEEQU_MAX_NUM_INSTANCES"] = equ_handle.XDFEEQU_MAX_NUM_INSTANCES
        equ_macro["XST_SUCCESS"] = equ_handle.XST_SUCCESS
        equ_macro["XST_FAILURE"] = equ_handle.XST_FAILURE
        equ_macro["XDFEEQU_NODE_NAME_MAX_LENGTH"] = equ_handle.XDFEEQU_NODE_NAME_MAX_LENGTH
        equ_macro["XDFEEQU_ANT_NUM_MAX"] = equ_handle.XDFEEQU_ANT_NUM_MAX
        equ_macro["XDFEEQU_CHANNEL_NUM"] = equ_handle.XDFEEQU_CHANNEL_NUM
        equ_macro["XDFEEQU_MAX_NUMBER_OF_UNITS_COMPLEX"] = equ_handle.XDFEEQU_MAX_NUMBER_OF_UNITS_COMPLEX
        equ_macro["XDFEEQU_MAX_NUMBER_OF_UNITS_REAL"] = equ_handle.XDFEEQU_MAX_NUMBER_OF_UNITS_REAL
        equ_macro["XDFEEQU_NUM_COEFF"] = equ_handle.XDFEEQU_NUM_COEFF
        equ_macro["XDFEEQU_DATAPATH_MODE_REAL"] = equ_handle.XDFEEQU_DATAPATH_MODE_REAL
        equ_macro["XDFEEQU_DATAPATH_MODE_COMPLEX"] = equ_handle.XDFEEQU_DATAPATH_MODE_COMPLEX
        equ_macro["XDFEEQU_DATAPATH_MODE_MATRIX"] = equ_handle.XDFEEQU_DATAPATH_MODE_MATRIX
        return equ_macro

    # Get enum API
    def GetEnum_XDfeEqu_StateId(self):
        """
        Return Dictionary equivalent of enum XDfeEqu_StateId

        :param : None
        :return: Dictionary equivalent of enum XDfeEqu_StateId
        """
        XDfeEqu_StateId = ffi.typeof("enum XDfeEqu_StateId").relements
        self.logger.debug(f"XDfeEqu_StateId = {json.dumps(XDfeEqu_StateId, indent=2)}")
        return XDfeEqu_StateId

    def GetEnum_metal_log_level(self):
        """
        Return Dictionary equivalent of enum metal_log_level

        :param : None
        :return: Dictionary equivalent of enum metal_log_level
        """
        metal_log_level = ffi.typeof("enum metal_log_level").relements
        self.logger.debug(f"metal_log_level = {json.dumps(metal_log_level, indent=2)}")
        return metal_log_level

    # Get structure  API
    def GetStruct_XDfeEqu_Version(self):
        """
        Return Dictionary equivalent of structure XDfeEqu_Version

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu_Version
        """
        XDfeEqu_Version_ptr = ffi.new("XDfeEqu_Version *")
        XDfeEqu_Version = cdata_to_py(XDfeEqu_Version_ptr[0])
        return XDfeEqu_Version

    def GetStruct_XDfeEqu_Trigger(self):
        """
        Return Dictionary equivalent of structure XDfeEqu_Trigger

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu_Trigger
        """
        XDfeEqu_Trigger_ptr = ffi.new("XDfeEqu_Trigger *")
        XDfeEqu_Trigger = cdata_to_py(XDfeEqu_Trigger_ptr[0])
        return XDfeEqu_Trigger

    def GetStruct_XDfeEqu_TriggerCfg(self):
        """
        Return Dictionary equivalent of structure XDfeEqu_TriggerCfg

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu_TriggerCfg
        """
        XDfeEqu_TriggerCfg_ptr = ffi.new("XDfeEqu_TriggerCfg *")
        XDfeEqu_TriggerCfg = cdata_to_py(XDfeEqu_TriggerCfg_ptr[0])
        return XDfeEqu_TriggerCfg

    def GetStruct_XDfeEqu_ModelParameters(self):
        """
        Return Dictionary equivalent of structure XDfeEqu_ModelParameters

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu_ModelParameters
        """
        XDfeEqu_ModelParameters_ptr = ffi.new("XDfeEqu_ModelParameters *")
        XDfeEqu_ModelParameters = cdata_to_py(XDfeEqu_ModelParameters_ptr[0])
        return XDfeEqu_ModelParameters

    def GetStruct_XDfeEqu_Cfg(self):
        """
        Return Dictionary equivalent of structure XDfeEqu_Cfg

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu_Cfg
        """
        XDfeEqu_Cfg_ptr = ffi.new("XDfeEqu_Cfg *")
        XDfeEqu_Cfg = cdata_to_py(XDfeEqu_Cfg_ptr[0])
        return XDfeEqu_Cfg

    def GetStruct_XDfeEqu_Coefficients(self):
        """
        Return Dictionary equivalent of structure XDfeEqu_Coefficients

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu_Coefficients
        """
        XDfeEqu_Coefficients_ptr = ffi.new("XDfeEqu_Coefficients *")
        XDfeEqu_Coefficients = cdata_to_py(XDfeEqu_Coefficients_ptr[0])
        return XDfeEqu_Coefficients

    def GetStruct_XDfeEqu_EqConfig(self):
        """
        Return Dictionary equivalent of structure XDfeEqu_EqConfig

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu_EqConfig
        """
        XDfeEqu_EqConfig_ptr = ffi.new("XDfeEqu_EqConfig *")
        XDfeEqu_EqConfig = cdata_to_py(XDfeEqu_EqConfig_ptr[0])
        return XDfeEqu_EqConfig

    def GetStruct_XDfeEqu_Status(self):
        """
        Return Dictionary equivalent of structure XDfeEqu_Status

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu_Status
        """
        XDfeEqu_Status_ptr = ffi.new("XDfeEqu_Status *")
        XDfeEqu_Status = cdata_to_py(XDfeEqu_Status_ptr[0])
        return XDfeEqu_Status

    def GetStruct_XDfeEqu_InterruptMask(self):
        """
        Return Dictionary equivalent of structure XDfeEqu_InterruptMask

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu_InterruptMask
        """
        XDfeEqu_InterruptMask_ptr = ffi.new("XDfeEqu_InterruptMask *")
        XDfeEqu_InterruptMask = cdata_to_py(XDfeEqu_InterruptMask_ptr[0])
        return XDfeEqu_InterruptMask

    def GetStruct_XDfeEqu_Config(self):
        """
        Return Dictionary equivalent of structure XDfeEqu_Config

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu_Config
        """
        XDfeEqu_Config_ptr = ffi.new("XDfeEqu_Config *")
        XDfeEqu_Config = cdata_to_py(XDfeEqu_Config_ptr[0])
        return XDfeEqu_Config

    def GetStruct_XDfeEqu(self):
        """
        Return Dictionary equivalent of structure XDfeEqu

        :param : None
        :return: Dictionary equivalent of structure XDfeEqu
        """
        XDfeEqu_ptr = ffi.new("XDfeEqu *")
        XDfeEqu = cdata_to_py(XDfeEqu_ptr[0])
        return XDfeEqu

    # System initialization API
    def XDfeEqu_InstanceInit(self, DeviceNodeName):
        """
        API initialises one instance of an Equalizer driver.
        Traverse "/sys/bus/platform/device" directory (in Linux), to find registered
        EQU device with the name DeviceNodeName. The first available slot in
        the instances array XDfeEqu_ChFilter[] will be taken as a DeviceNodeName
        object.

        :param DeviceNodeName: device node name.
        :return: ret - 0 on success, 1 on failure
                 device_id - integer handle to initialized instance
                 DeviceNodeNameRet - device node name returned from driver
                                     which will be same as the passed value

        """
        DeviceNodeName = extract_b64_encoded_string(DeviceNodeName)
        cptrDeviceNodeName = ffi.new("char[128]", DeviceNodeName)
        self.logger.info(f"XDfeEqu_InstanceInit DeviceNodeName = {DeviceNodeName}")
        # Pass the device node name to the driver
        xequ = equ_handle.XDfeEqu_InstanceInit(cptrDeviceNodeName)
        self.logger.info(f"The initalized pointer value = {xequ}")
        DeviceNodeNameRet = cdata_string_to_py(cptrDeviceNodeName);
        if xequ == ffi.NULL:
            # if driver returns NULL, pass it as error to user
            ret = 1        
            self.logger.error(f"The pointer returned for {DeviceNodeName} is NULL")
            device_id = 0xFFFFFFFF
        else:
            # otherwise treat as success
            ret = 0        
            # In equ_dict dictionary, device id is the key.
            # device node name and instance pointer from driver are values as list
            # check if the device node name is already present in dictionary
            found, device_id = getkey_from_listbeginvalue(self.equ_dict, DeviceNodeName)
            if 1 == found:
                # if found, update the device id key with values
                self.equ_dict[device_id][0] = DeviceNodeNameRet 
                self.equ_dict[device_id][1] = xequ                            
            else:
                # if not found, make new device_id from class variable
                # and increment the class variable
                device_id = self.device_id
                self.device_id += 1
                # add new key and values to the dictionary
                self.equ_dict[device_id] = [DeviceNodeNameRet, xequ]            
        self.logger.info(f"ret = {ret}, device_id = {device_id}, DeviceNodeNameRet = {DeviceNodeNameRet}")
        return ret, device_id, DeviceNodeNameRet

    def XDfeEqu_InstanceClose(self, device_id):
        """
        API closes the instance of an Equalizer driver and moves the state machine
        to a Not Ready state.

        :param device_id: id of the opened device.
        :return: None
        """
        self.logger.info(f"XDfeEqu_InstanceClose({device_id})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_InstanceClose(xequ)
        if device_id in self.equ_dict:
            del self.equ_dict[device_id]        
        return

    # Register access API
    def XDfeEqu_WriteReg(self, device_id, addr_offset, data):
        """
        Writes value to register in an Equalizer instance.

        :param device_id: id of the opened device.
        :param AddrOffset: is address offset relative to instance base address.
        :param Data: is value to be written.
        :return: None
        """
        self.logger.debug(f"XDfeEqu_WriteReg({device_id}, {hex(addr_offset)}, {hex(data)})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_WriteReg(xequ, addr_offset, data)
        return

    def XDfeEqu_ReadReg(self, device_id, addr_offset):
        """
        Reads a value from the register using an Equalizer instance.

        :param device_id: id of the opened device.
        :param AddrOffset: is address offset relative to instance base address.
        :return: Register value.
        """
        self.logger.debug(f"ret = XDfeEqu_ReadReg({device_id}, {hex(addr_offset)})")
        xequ = self.equ_dict[device_id][1]
        regval = equ_handle.XDfeEqu_ReadReg(xequ, addr_offset)
        self.logger.debug(f"regval = {regval}")
        return regval

    # DFE Equalizer component initialization API
    def XDfeEqu_Reset(self, device_id):
        """
        Resets Equalizer and puts block into a reset state.

        :param device_id: id of the opened device.
        :return: None
        """
        self.logger.debug(f"XDfeEqu_Reset({device_id})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_Reset(xequ)
        return

    def XDfeEqu_Configure(self, device_id, Cfg):
        """
        Reads configuration from device tree/xparameters.h and IP registers.
        Removes S/W reset and moves the state machine to a Configured state.

        :param device_id: id of the opened device.
        :param Cfg: device config container.
        :return: Cfg: device config container.
        """
        Cfg_ptr = ffi.new("XDfeEqu_Cfg *", Cfg)
        self.logger.debug(f"XDfeEqu_Configure({device_id}, {json.dumps(Cfg, indent=2)})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_Configure(xequ, Cfg_ptr)
        Cfg = cdata_to_py(Cfg_ptr[0])
        self.logger.debug(f"Cfg = {json.dumps(Cfg, indent=2)}")
        return Cfg

    def XDfeEqu_Initialize(self, device_id, Config):
        """
        DFE Equalizer driver one time initialisation and moves the state machine to
        an Initialised state.

        :param device_id: id of the opened device.
        :param Config: configuration data container.
        :return: Config: configuration data container.
        """
        Config_ptr = ffi.new("XDfeEqu_EqConfig *", Config)
        self.logger.debug(f"XDfeEqu_Initialize({device_id}, {json.dumps(Config, indent=2)})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_Initialize(xequ, Config_ptr)
        Config = cdata_to_py(Config_ptr[0])
        self.logger.debug(f"Init = {json.dumps(Config, indent=2)}")
        return Config

    def XDfeEqu_Activate(self, device_id, EnableLowPower):
        """
        Activates channel Equalizer moves the state machine to an Activated state.

        :param device_id: id of the opened device.
        :param EnableLowPower: flag indicating low power.
        :return: None
        """
        self.logger.debug(f"XDfeEqu_Activate({device_id}, {EnableLowPower})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_Activate(xequ, EnableLowPower)
        return

    def XDfeEqu_Deactivate(self, device_id):
        """
        Deactivates Equalizer and moves the state machine to Initialised state.

        :param device_id: id of the opened device.
        :return: None
        """
        self.logger.debug(f"XDfeEqu_Deactivate({device_id})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_Deactivate(xequ)
        return

    def XDfeEqu_GetStateID(self, device_id):
        """
        Gets a state machine state id. The state id is returned as a string.
        The enum is mapped to a dictionary in python.

        :param device_id: id of the opened device.
        :return: StateId: State machine StateID as a string
        """
        self.logger.debug(f"XDfeEqu_GetStateID({device_id})")
        xequ = self.equ_dict[device_id][1]
        StateId = equ_handle.XDfeEqu_GetStateID(xequ)
        StateId_edict = ffi.typeof("XDfeEqu_StateId").elements
        self.logger.debug(f"Return value = {json.dumps(StateId, indent=2)}")
        return StateId_edict[StateId]

    # User APIs
    def XDfeEqu_Update(self, device_id, Config):
        """
        The software first sets bits 7:4 of the Next Control register (0x24) with
        the values in Config.Real_Datapath_Set, Config.Im_Datapath_Set.
        In real mode bits, 5:4 are set to the value held in
        Config.Real_Datapath_Set.
        Bits 7:6 are set to the value held in Config.Real_Datapath_Set.
        In complex mode bits, 5:4 are set to the value held in value
        Config.Real_Datapath_Set. Bits 7:6 are set to the value held in
        Config.Real_Datapath_Set plus 1.
        In matrix mode bits 5:4 are set to the value held in
        Config.Real_Datapath_Set. Bits 7:6 are set to the value held in
        Config.Im_Datapath_Set.
        The software sets bit 1 depending on Config.Flush.
        The software then sets the _Enable bit in the Next Control Trigger
        Source register (0x28) high.

        :param device_id: id of the opened device.
        :param Config: configuration container
        :return: Config: configuration container
        """
        Config_ptr = ffi.new("XDfeEqu_EqConfig *", Config)
        self.logger.debug(f"XDfeEqu_Update({device_id}, {json.dumps(Config, indent=2)})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_Update(xequ, Config_ptr)
        Config = cdata_to_py(Config_ptr[0])
        self.logger.debug(f"Config = {json.dumps(Config, indent=2)}")
        return Config

    def XDfeEqu_GetTriggersCfg(self, device_id):
        """
        Returns current trigger configuration.

        :param device_id: id of the opened device.
        :return: TriggerCfg: is a trigger configuration container.
        """
        TriggerCfg_ptr = ffi.new("XDfeEqu_TriggerCfg *")
        self.logger.debug(f"XDfeEqu_GetTriggersCfg({device_id})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_GetTriggersCfg(xequ, TriggerCfg_ptr)
        TriggerCfg = cdata_to_py(TriggerCfg_ptr[0])
        self.logger.debug(f"TriggerCfg = {json.dumps(TriggerCfg, indent=2)}")
        return TriggerCfg

    def XDfeEqu_SetTriggersCfg(self, device_id, TriggerCfg):
        """
        Set trigger configuration.

        :param device_id: id of the opened device.
        :param TriggerCfg: is a triger configuration container.
        :return: TriggerCfg: is a triger configuration container.
        """
        TriggerCfg_ptr = ffi.new("XDfeEqu_TriggerCfg *", TriggerCfg)
        self.logger.debug(f"XDfeEqu_SetTriggersCfg({device_id}, {json.dumps(TriggerCfg, indent=2)})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_SetTriggersCfg(xequ, TriggerCfg_ptr)
        TriggerCfg = cdata_to_py(TriggerCfg_ptr[0])
        self.logger.debug(f"The return value TriggerCfg = {json.dumps(TriggerCfg, indent=2)}")
        return TriggerCfg

    def XDfeEqu_LoadCoefficients(self, device_id, ChannelField, Mode, Shift, EqCoeffs):
        """
        Sets Equalizer filter coefficients in Real, Complex or Matrix mode.

        :param device_id: id of the opened device.
        :param ChannelField: is a flag in which bits indicate the channel is enabled.
        :param Mode: is an equalizer mode.
        :param Shift: is a coefficient shift value.
        :param EqCoeffs: is equalizer coefficients container.
        :return: EqCoeffs: is equalizer coefficients container.
        """
        EqCoeffs_ptr = ffi.new("XDfeEqu_Coefficients *", EqCoeffs)
        self.logger.debug(f"XDfeEqu_LoadCoefficients({device_id}, {ChannelField}, {Mode}, {Shift}, {json.dumps(EqCoeffs, indent=2)})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_LoadCoefficients(xequ, ChannelField, Mode, Shift, EqCoeffs_ptr)
        EqCoeffs = cdata_to_py(EqCoeffs_ptr[0])
        self.logger.debug(f"EqCoeffs = {EqCoeffs}")
        return EqCoeffs

    def XDfeEqu_GetEventStatus(self, device_id):
        """
        Gets Equalizer event status

        :param device_id: id of the opened device.
	    :return: Status: event status
        """
        Status_ptr = ffi.new("XDfeEqu_Status *")
        self.logger.debug(f"XDfeEqu_GetEventStatus({device_id})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_GetEventStatus(xequ, Status_ptr)
        Status = cdata_to_py(Status_ptr[0])
        self.logger.debug(f"Status = {json.dumps(Status, indent=2)}")
        return Status

    def XDfeEqu_ClearEventStatus(self, device_id, Status):
        """
        Clears Equalizer status. The channel status will be cleared for any IStatus
        or QStatus not equal 0.
        Note: The Status registers are only present for a given channel when it
              is present in the configured IP. The number of channels present is
              given by CONFIG.NUM_CHANNELS.        

        :param device_id: id of the opened device.
        :param Status: event status.
        :return: None
        """
        Status_ptr = ffi.new("XDfeEqu_Status *", Status)
        self.logger.debug(f"XDfeEqu_ClearEventStatus({device_id}, {json.dumps(Status, indent=2)})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_ClearEventStatus(xequ, Status_ptr)
        return

    def XDfeEqu_SetInterruptMask(self, device_id, InterruptMask):
        """
        Enables an Equalizer status for channel ID.
        Note: The Status Mask registers are only present for a given channel when
               it is present in the configured IP. The number of channels present
               is given by CONFIG.NUM_CHANNELS.
               
        :param device_id: id of the opened device.
        :param InterruptMask: interrupt mask value
        :return: None
        """
        InterruptMask_ptr = ffi.new("XDfeEqu_InterruptMask *", InterruptMask)
        self.logger.debug(f"XDfeEqu_SetInterruptMask({device_id}, {json.dumps(InterruptMask, indent=2)})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_SetInterruptMask(xequ, InterruptMask_ptr)
        return

    def XDfeEqu_GetInterruptMask(self, device_id):
        """
        Gets interrupt mask.

        :param device_id: id of the opened device.
        :return: InterruptMask: Equalizer interrupt mask container.
        """
        InterruptMask_ptr = ffi.new("XDfeEqu_InterruptMask *")
        self.logger.debug(f"XDfeEqu_GetInterruptMask({device_id})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_GetInterruptMask(xequ, InterruptMask_ptr)
        InterruptMask = cdata_to_py(InterruptMask_ptr[0])
        self.logger.debug(f"InterruptMask = {json.dumps(InterruptMask, indent=2)}")
        return InterruptMask

    def XDfeEqu_GetActiveSets(self, device_id):
        """
        Gets used coefficients settings.

        :param device_id: id of the opened device.
        :return: RealSet: Real value
                 ImagSet: Imaginary value
        """
        RealSet_ptr = ffi.new("u32 *")
        ImagSet_ptr = ffi.new("u32 *")
        self.logger.debug(f"XDfeEqu_GetActiveSets({device_id})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_GetActiveSets(xequ, RealSet_ptr, ImagSet_ptr)
        RealSet = RealSet_ptr[0]
        ImagSet = ImagSet_ptr[0]
        self.logger.debug(f"RealSet = {json.dumps(RealSet, indent=2)}, ImagSet = {json.dumps(ImagSet, indent=2)}")
        return RealSet, ImagSet

    def XDfeEqu_SetTUserDelay(self, device_id, Delay):
        """
        Sets the delay, which will be added to TUSER and TLAST (delay matched
        through the IP).

        :param device_id: id of the opened device.
        :param Delay: requested delay variable.
        :return: None
        """
        self.logger.debug(f"XDfeEqu_SetTUserDelay({device_id}, {Delay})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_SetTUserDelay(xequ, Delay)
        return

    def XDfeEqu_GetTUserDelay(self, device_id):
        """
        Reads the delay, which will be added to TUSER and TLAST (delay matched
        through the IP).

        :param device_id: id of the opened device.
        :return: ret: Delay value
        """
        self.logger.debug(f"XDfeEqu_GetTUserDelay({device_id})")
        xequ = self.equ_dict[device_id][1]
        ret = equ_handle.XDfeEqu_GetTUserDelay(xequ)
        self.logger.debug(f"ret = {ret}")
        return ret

    def XDfeEqu_GetTDataDelay(self, device_id, Tap):
        """
        Returns CONFIG.DATA_LATENCY.VALUE + tap, where the tap is between 0
        and 23 in real mode and between 0 and 11 in complex/matrix mode.

        :param device_id: id of the opened device.
        :param Tap: is a tap variable.
        :return: ret: data latency value.
        """
        self.logger.debug(f"XDfeEqu_GetTDataDelay({device_id}, {Tap})")
        xequ = self.equ_dict[device_id][1]
        ret = equ_handle.XDfeEqu_GetTDataDelay(xequ, Tap)
        self.logger.debug(f"ret = {ret}")
        return ret

    def XDfeEqu_GetVersions(self, device_id):
        """
        This API is used to get the driver version.

        :param device_id: id of the opened device.
        :return: SwVersion: driver version numbers.
                 HwVersion: HW version numbers.
        """
        SwVersion_ptr = ffi.new("XDfeEqu_Version *")
        HwVersion_ptr = ffi.new("XDfeEqu_Version *")
        self.logger.debug(f"XDfeEqu_GetVersions({device_id})")
        xequ = self.equ_dict[device_id][1]
        equ_handle.XDfeEqu_GetVersions(xequ, SwVersion_ptr, HwVersion_ptr)
        SwVersion = cdata_to_py(SwVersion_ptr[0])
        HwVersion = cdata_to_py(HwVersion_ptr[0])
        self.logger.debug(f"SwVersion = {json.dumps(SwVersion, indent=2)}")
        self.logger.debug(f"HwVersion = {json.dumps(HwVersion, indent=2)}")
        return SwVersion, HwVersion

    def __del__(self):
        self.logger.info("Inside EQU Destructor")



# Copyright (C) 2021-2022 Xilinx, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

__author__ = "Anish Kadamathikuttiyil Karthikeyan Pillai"
__copyright__ = "Copyright 2021, Xilinx"

import Pyro4
import logging
import base64
from utils import convert_to_bytearray, convert_to_list
import serpent
from axi_memmap_c import AXI_MEMMAP_C


class AXI_MEMMAP(object):
    memmap_c = None
    def __init__(self):
        self.memmap_c = AXI_MEMMAP_C()
        logging.info("Inside AXI MEMMAP Constructor")
        pass

    def axi_read_words(self, address, num_words, network_order):
        """
        Read num_words from address.

        :param address: address to read from
        :param num_words: number of words to read
        :param network_order: Apply network byte ordering if network_order!=0
        :return: ret: whether success or failure
                 buf: the read data

        """
        ret, buf = self.memmap_c.axi_read_words(address, num_words, network_order)
        buf = convert_to_bytearray(buf)
        if ret == 0:
            logging.info("axi_memmap_c.axi_read_words returned success")
        else:
            logging.info("axi_memmap_c.axi_read_words returned failure ret = ", ret)
        return ret, buf

    def axi_write_words(self, address, num_words, buf, network_order):
        """
        Write num_words to address.

        :param address: address to write
        :param num_words: number of words to write
        :param buffer: data to write
        :param network_order: Apply network byte ordering if network_order!=0
        :return: ret: whether success or failure

        """
        if Pyro4.config.SERIALIZER == "serpent" and type(buf) is dict:
            buf = serpent.tobytes(buf)  # in case of serpent encoded bytes
        buf = convert_to_list(buf[0:num_words*4])
        ret = self.memmap_c.axi_write_words(address, num_words, buf,  network_order)
        if ret == 0:
            logging.info("axi_memmap_c.axi_write_words returned success")
        else:
            logging.info("axi_memmap_c.axi_write_words returned failure ret = ", ret)
        return ret

    def __del__(self):
        logging.info("Inside AXI MEMMAP Destructor")




# Copyright (C) 2023-2025 Advanced Micro Devices, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

__author__ = "Salih Erim"
__copyright__ = "Copyright 2023-2025, Advanced Micro Devices, Inc."

import os
import sys
import logging
import dbm.dumb

RAFT_DIR = '/usr/share/raft/'
sys.path.append(RAFT_DIR + 'xserver/utils')
sys.path.append(RAFT_DIR + 'xserver/raft_services/power_management/devices')

from pmic import PMIC
from devices.stats import Stats
from devices.sysmon import Sysmon
from utils import get_python_log_levels
from pm_types import *

class HardwareError(Exception):
    pass

class PM(object):
    logger = None
    domains = []
    sensors = []
    voltages = []
    scales = None
    pmic = None
    temps = []
    status = None
    pdi_file = ""

    def __init__(self, json_data, board_name, eeprom):
        self.logger = self.GetLogger()
        self.boardeeprom = eeprom
        self.board_name = board_name
        board_data = json_data[board_name]
        self.pdi_file = None

        if 'FEATURE' in board_data:
            self.feature_list = board_data['FEATURE']['List']

        self.unit_config = {
            "voltage": {
                "base_unit": "V",
                "available_scales": {
                    "base": 1,
                    "milli": 1e-3,
                }
            },
            "current": {
                "base_unit": "A",
                "available_scales": {
                    "base": 1,
                    "milli": 1e-3,
                    "micro": 1e-6,
                }
            },
            "power": {
                "base_unit": "W",
                "available_scales": {
                    "base": 1,
                    "milli": 1e-3,
                }
            },
            "temperature": {
                "base_unit": "C",
                "available_scales": {
                    "base": 1,
                    "milli": 1e-3,
                }
            }
        }
        self.scales = {param: "base" for param in self.unit_config}

        if 'powerdomain' in self.feature_list:
            if 'POWER DOMAIN' in board_data:
                for key, val in board_data['POWER DOMAIN'].items():
                    temp_d = Domain(**val)
                    for index, railname in enumerate(temp_d.railnames):
                        for k, v in board_data['POWER_SENSORS'].items():
                            if k == railname:
                                if temp_d.railnames[index] != v['Name']:
                                    temp_d.railnames[index] = v['Name']
                    self.domains.append(temp_d)

        if 'power' in self.feature_list:
            if 'POWER_SENSORS' in board_data:
                for k, v in board_data['POWER_SENSORS'].items():
                    temp_r = Rails(**v)
                    self.sensors.append(temp_r)

        if 'voltage' in self.feature_list:
            if 'VOLTAGE' in board_data:
                for k, v in board_data['VOLTAGE'].items():
                    tempVolt = Rails(**v)
                    self.voltages.append(tempVolt)

        if 'temp' in self.feature_list:
            if 'Temperature' in board_data:
                if "versal-isa" in board_data['Temperature']['Sensor']:
                    try:
                        temp_s = {
                            "Name" : board_data['Temperature']['Name'],
                            "Sensor" : Sysmon(None, None, Sysmon_Device_Type.SYSFS)
                        }
                        if temp_s is not None:
                            self.temps.append(temp_s)
                    except Exception as e:
                        print(e)
                elif "i2c" in board_data['Temperature']['Sensor']:
                    tokens = board_data['Temperature']['Sensor'].split("-")
                    address = tokens[3]
                    device_path = "/dev/i2c-" + tokens[2]
                    try:
                        temp_s = {
                            "Name" : board_data['Temperature']['Name'],
                            "Sensor" : Sysmon(address,device_path, Sysmon_Device_Type.I2C)
                            }
                        if temp_s is not None:
                            self.temps.append(temp_s)
                    except Exception as e:
                        print(e)

        if 'Boot Config' in board_data:
            self.pdi_file = board_data['Boot Config']['PDI']

        try:
            self.status = Stats()
            if self.status is None:
                self.logger.error(f"InitStats failed.")
        except Exception as e:
            self.logger.error(f"InitStats failed. ({e})")
            self.exit_program()

        try:
            self.pmic = PMIC(self.domains, self.sensors, self.voltages)
        except Exception as e:
            self.logger.error(f"InitPmic failed. ({e})")
            self.exit_program()

        if self.pmic is None:
            self.logger.error(f"PMIC not found.")
            self.exit_program()
        self.logger.info("Inside PM Constructor")

    @staticmethod
    def exit_program():
        logging.critical("CRITICAL ERROR: Unable to Run Pyro Server.\n")
        sys.exit(1)

    @staticmethod
    def GetLogger():
        """
        Static method to get the logger for the class.
        Default loglevel is set inside this class

        :return: logger

        """
        log_level = logging.ERROR
        logging.basicConfig(format="%(levelname)s:%(message)s")
        logger = logging.getLogger(__name__)
        try:
            handler_set_check = getattr(logger, 'handler_set')
        except AttributeError:
            handler_set_check = False
        if not handler_set_check:
            logger.setLevel(log_level)
            logger.handler_set = True
            logger.disabled = False
        return logger

    # Log level
    def GetPythonLogLevels(self):
        """
        Return the logging levels supported by logging library in python

        :param : None
        :return: Dictionary showing the log levels supported by logging library
        """
        return get_python_log_levels()

    def SetServerLogLevel(self, PythonLogLevel):
        """
        Set the python log level to the given level

        :param : Log level to set
        :return: None
        """
        self.logger.debug(f"PythonLogLevel = {PythonLogLevel}")
        LogLevelsDict = get_python_log_levels()
        if PythonLogLevel == LogLevelsDict["DEBUG"]:
            self.logger.setLevel(logging.DEBUG)
        elif PythonLogLevel == LogLevelsDict["INFO"]:
            self.logger.setLevel(logging.INFO)
        elif PythonLogLevel == LogLevelsDict["WARNING"]:
            self.logger.setLevel(logging.WARNING)
        elif PythonLogLevel == LogLevelsDict["ERROR"]:
            self.logger.setLevel(logging.ERROR)
        else:
            self.logger.setLevel(logging.CRITICAL)
        return

    def handle_request(self, process_function, *args, **kwargs):
        try:
            caller_name = sys._getframe(1).f_code.co_name.lower()
            if 'voltage' in caller_name:
                if 'voltage' not in self.feature_list:
                    raise Exception(f"Unsupported feature: voltage")
            if 'regulator' in caller_name:
                if 'voltage' not in self.feature_list:
                    raise Exception(f"Unsupported feature: voltage")
            if 'power' in caller_name:
                if 'power' not in self.feature_list:
                    raise Exception(f"Unsupported feature: power")
            if 'temp' in caller_name:
                if 'temp' not in self.feature_list:
                    raise Exception(f"Unsupported feature: temperature")
            if 'domain' in caller_name:
                if 'powerdomain' not in self.feature_list:
                    raise Exception(f"Unsupported feature: powerdomain")

            data = process_function(*args, **kwargs)
            return {
                "status": "success",
                "data": data,
                "message": "Operation completed successfully."
            }
        except HardwareError as e:
            return {
                "status": "failure",
                "data": None,
                "message": f"Hardware error: {str(e)}"
            }
        except ValueError as e:
            self.logger.error(f'{sys._getframe(1).f_code.co_name}{args} : {str(e)}')
            return {
                "status": "failure",
                "data": None,
                "message": f"Value error: {str(e)}"
            }
        except Exception as e:
            self.logger.error(f'{sys._getframe(1).f_code.co_name}{args} : {str(e)}')
            return {
                "status": "failure",
                "data": None,
                "message": f"Unexpected error: {str(e)}"
            }

    # Call APIs
    def GetBoardInfo(self):
        self.logger.info(f"GetBoardInfo()")
        return self.handle_request(self._get_board_info)

    def ListFeature(self):
        self.logger.info(f"GetBoardInfo()")
        return self.handle_request(self._list_feature)


    def ListPowerDomains(self):
        self.logger.info(f"ListPowerSensors()")
        return self.handle_request(self._list_power_domains)

    def ListRailsOfDomain(self, domain_name):
        self.logger.info(f"ListRailsOfDomain()")
        return self.handle_request(self._list_rails_of_domain, domain_name)

    def GetValueOfDomain(self, domain_name):
        self.logger.info(f"GetValueOfDomain()")
        return self.handle_request(self._get_value_of_domain, domain_name)

    def GetPowersAll(self):
        self.logger.info(f"GetPowersAll()")
        return self.handle_request(self._get_powers_all)

    def GetValuesAll(self):
        self.logger.info(f"GetValuesAll()")
        return self.handle_request(self._get_values_all)

    """
    Power Sensor APIs
    """
    def ListPowerSensors(self):
        self.logger.info(f"ListPowerSensors()")
        return self.handle_request(self._list_power_sensor)

    def GetPowerSensor(self, sensor_name):
        self.logger.info(f"GetPowerSensor({sensor_name})")
        return self.handle_request(self._get_power_sensor, sensor_name)

    def GetCalPowerSensor(self, sensor_name):
        self.logger.info(f"GetCalPowerSensor({sensor_name})")
        return self.handle_request(self._get_cal_power_sensor, sensor_name)

    def GetPowerSensorConf(self, sensor_name):
        self.logger.info(f"GetPowerSensorConf({sensor_name})")
        return self.handle_request(self._get_power_sensor_conf, sensor_name)

    def SetPowerSensorConf(self, sensor_name, conf):
        self.logger.info(f"SetPowerSensorConf({sensor_name, conf})")
        return self.handle_request(self._set_power_sensor_conf, sensor_name, conf)

    """
    Voltage Regulator APIs
    """
    def ListVoltages(self):
        self.logger.info(f"ListVoltages")
        return self.handle_request(self._list_voltages)

    def EnableVoltage(self, voltage_name):
        self.logger.info(f"EnableVoltage({voltage_name})")
        return self.handle_request(self._enable_voltage, voltage_name)

    def DisableVoltage(self, voltage_name):
        self.logger.info(f"DisableVoltage({voltage_name})")
        return self.handle_request(self._disable_voltage, voltage_name)

    def GetRegulator(self, voltage_name):
        self.logger.info(f"GetRegulator({voltage_name})")
        return self.handle_request(self._get_regulator, voltage_name)

    def GetVoltage(self, voltage_name):
        self.logger.info(f"GetVoltage({voltage_name})")
        return self.handle_request(self._get_voltage, voltage_name)

    def SetVoltage(self, voltage_name, new_value):
        self.logger.info(f"SetVoltage({voltage_name}, {new_value})")
        return self.handle_request(self._set_voltage, voltage_name, new_value)

    def SetBootVoltage(self, voltage_name, new_value):
        self.logger.info(f"SetBootVoltage({voltage_name}, {new_value})")
        return self.handle_request(self._set_boot_voltage, voltage_name, new_value)

    def RestoreVoltage(self, voltage_name):
        self.logger.info(f"SetVoltage({voltage_name})")
        return self.handle_request(self._restore_voltage, voltage_name)

    def ListTemperatures(self):
        self.logger.info(f"ListTemperatures()")
        return self.handle_request(self._list_temperatures)

    def GetTemperature(self, temp_name):
        self.logger.info(f"GetSysmonTemperature()")
        return self.handle_request(self._get_temperature, temp_name)

    def ListUnits(self):
        self.logger.info(f"ListUnits()")
        return self.handle_request(self._list_units)

    def GetUnit(self, quantity):
        self.logger.info(f"GetUnit()")
        return self.handle_request(self._get_unit, quantity)

    def ListAvailableScales(self, quantity):
        self.logger.info(f"ListAvailableScales()")
        return self.handle_request(self._list_available_scales, quantity)

    def SetScale(self, quantity, unit):
        self.logger.info(f"SetUnit()")
        return self.handle_request(self._set_scale, quantity, unit)

    def _list_units(self):
        if self.scales is None:
            raise Exception(f'Unsupported feature: Enable "Units" in config')
        return {
            param: {
                "current_scale": self.scales[param],
                "current_unit": self._get_unit(param),
                "available_scales": list(self.unit_config[param]["available_scales"].keys()),
                "base_unit": self.unit_config[param]["base_unit"]
            }
            for param in self.unit_config
        }

    def _get_unit(self, quantity):
        if self.scales is None:
            raise Exception(f'Unsupported feature: Enable "Units" in config')
        if quantity not in self.scales:
            raise ValueError(f"Unsupported quantity: {quantity}")
        scale = self.scales[quantity]
        base_unit = self.unit_config[quantity]["base_unit"]

        # Define prefixes for scales
        scale_prefix_map = {
            "base": "",
            "milli": "m",
            "micro": "µ",
            "kilo": "k",
        }

        # Return the correctly formatted unit
        prefix = scale_prefix_map.get(scale, "")
        return f"{prefix}{base_unit}"

    def _set_scale(self, quantity, scale):
        if self.scales is None:
            raise Exception(f'Unsupported feature: Enable "Units" in config')
        if quantity not in self.unit_config:
            raise ValueError(f"Unsupported quantity: {quantity}")
        if scale not in self.unit_config[quantity]["available_scales"]:
            raise ValueError(f"Unsupported scale for {quantity}: {scale}")
        self.scales[quantity] = scale
    
    def _list_available_scales(self, quantity):
        if self.scales is None:
            raise Exception(f'Unsupported feature: Enable "Units" in config')
        if quantity not in self.unit_config:
            raise ValueError(f"Unsupported value: {quantity}")
        return list(self.unit_config[quantity]["available_scales"].keys())

    def _scale(self, value, quantity):
        if value is None:
            return value
        if self.scales is None:
            return value
        if quantity not in self.scales:
            raise ValueError(f"Unsupported value: {quantity}")
        scale = self.scales[quantity]
        factor = self.unit_config[quantity]["available_scales"][scale]
        return round((value / factor), 3)

    def _unscale(self, value, quantity):
        if value is None:
            return value
        if self.scales is None:
            return value
        if quantity not in self.scales:
            raise ValueError(f"Unsupported value: {quantity}")
        scale = self.scales[quantity]
        factor = self.unit_config[quantity]["available_scales"][scale]
        return round((value * factor), 3)

    ## Process functions
    def _get_board_info(self):
        if self.boardeeprom is None:
            raise ValueError(f'Board Eeprom undefined')
        else:
            return self.pmic.GetBoardInfo(self.boardeeprom, self.pdi_file)

    def _list_feature(self):
        if len(self.feature_list) == 0:
            raise ValueError(f"Feature list is empty")
        return self.feature_list

    def __find_domain(self, domain_name):
        temp_d = None
        for d in self.domains:
            if d.name == domain_name:
                temp_d = d
        return temp_d

    def _list_power_domains(self):
        data = []
        for d in self.domains:
            data.append(d.name)
        if len(data) == 0:
            raise Exception(f'Power domains list is empty')
        return data

    def _list_rails_of_domain(self, domain_name):
        data = []
        domain = self.__find_domain(domain_name)
        if domain is None:
            raise ValueError(f'{domain_name} does not exists')
        else:
            for railname in domain.railnames:
                data.append(railname)
            if len(data) == 0:
                raise Exception(f'Rails list is empty')
        return data

    def _get_value_of_domain(self, domain_name):
        total_power = 0.0
        data = {}
        domain = self.__find_domain(domain_name)
        if domain is None:
            raise ValueError(f'{domain_name} does not exists')
        else:
            data[domain_name] = {}
            data[domain_name]['Rails'] = []
            for railname in domain.railnames:
                ps = self.__find_power_sensor(railname)
                if ps is None:
                    raise Exception(f'{domain_name} sensor is not defined')
                else:
                    v, i, p = self.pmic.GetSensorValues(ps._sensor)
                    rail_value = {
                        railname: {
                        'Voltage': self._scale(v, "voltage"),
                        'Current': self._scale(i, "current"),
                        'Power': self._scale(p, "power")
                        }
                    }
                    data[domain_name]['Rails'].append(rail_value)
                    total_power += self._scale(rail_value[railname]['Power'], "power")
            data[domain_name]['Total Power'] = round(total_power, 3)
        return data

    def _get_power_value_of_domain(self, domain_name):
        domain_power = 0.0
        data = {}
        domain = self.__find_domain(domain_name)
        if domain is None:
            raise ValueError(f'{domain_name} does not exists')
        else:
            data[domain_name] = {}
            domain_power = 0.0
            for railname in domain.railnames:
                ps = self.__find_power_sensor(railname)
                if ps is None:
                    raise Exception(f'{railname} sensor is not defined')
                else:
                    _, _, p = self.pmic.GetSensorValues(ps._sensor)
                    domain_power += self._scale(p, "power")
            data[domain_name]['Power'] = domain_power
        return data

    def _get_powers_all(self):
        data = {}
        data[self.board_name] = {}
        data[self.board_name]['Power Domains'] = []
        total_power = 0.0
        for domain in self.domains:
            temp_p = self._get_power_value_of_domain(domain.name)
            total_power += self._scale(temp_p[domain.name]['Power'], "power")
            data[self.board_name]['Power Domains'].append(temp_p)
        data[self.board_name]['Total Power'] = round(total_power, 3)
        return data

    def _get_values_all(self):
        data = {}
        data[self.board_name] = []
        for domain in self.domains:
            data[self.board_name].append(self._get_value_of_domain(domain.name))
        return data

    def __find_power_sensor(self, sensor_name):
        temp_s = None
        for ps in self.pmic.power_sensors:
            if ps.name == sensor_name:
                temp_s = ps
        return temp_s

    def _list_power_sensor(self):
        data = []
        if len(self.pmic.power_sensors) == 0:
            raise ValueError(f"Power Sensor list is empty")
        for s in self.pmic.power_sensors:
            temp = {
                s.name : {
                    "Device" : s.part_name
                }
            }
            data.append(temp)
        return data

    def _get_power_sensor(self, sensor_name):
        ps = self.__find_power_sensor(sensor_name)
        if ps is None:
            raise ValueError(f'{sensor_name} does not exits')
        else:
            if ps._sensor is None:
                raise Exception(f'{sensor_name} sensor is not defined')
            else:
                v, i, p = self.pmic.GetSensorValues(ps._sensor)
                data = {
                    'Voltage' : self._scale(v, "voltage"),
                    'Current' : self._scale(i, "current"),
                    'Power' : self._scale(p, "power")
                }
        return data

    def _get_cal_power_sensor(self, sensor_name):
        ps = self.__find_power_sensor(sensor_name)
        if ps is None:
            raise ValueError(f'{sensor_name} does not exits')
        else:
            if ps._sensor is None:
                raise Exception(f'{sensor_name} sensor is not defined')
            else:
                v, i, p = self.pmic.GetSensorValues(ps._sensor)
                data = {
                    'Voltage' : self._scale(v, "voltage"),
                    'Current' : self._scale(i, "current"),
                    'Power' : self._scale(p, "power")
                }
        return data

    def _get_power_sensor_conf(self, sensor_name):
        ps = self.__find_power_sensor(sensor_name)
        if ps is None:
            raise ValueError(f'{sensor_name} does not exits')
        else:
            if ps._sensor is None:
                raise Exception(f'{sensor_name} sensor is not defined')
        registers = self.pmic.GetPowerSensorConf(ps._sensor)
        data = {}
        match ps.part_name:
            case "INA226":
                data['Configuration'] = '0x{0:04x}'.format(registers[0])
                data['Shunt_Voltage'] = '0x{0:04x}'.format(registers[1])
                data['Bus_Voltage'] =   '0x{0:04x}'.format(registers[2])
                data['Power'] = '0x{0:04x}'.format(registers[3])
                data['Current'] = '0x{0:04x}'.format(registers[4])
                data['Calibration'] = '0x{0:04x}'.format(registers[5])
                data['Mask_Enable'] = '0x{0:04x}'.format(registers[6])
                data['Alert_Limit'] = '0x{0:04x}'.format(registers[7])
            case "INA700" | "INA745A" | "INA745B":
                data['Configuration'] = '0x{0:04x}'.format(registers[0])
                data['ADC_Configuration'] = '0x{0:04x}'.format(registers[1])
                data['Bus_Voltage'] =   '0x{0:04x}'.format(registers[2])
                data['Die_Temp'] = '0x{0:04x}'.format(registers[3])
                data['Current'] = '0x{0:04x}'.format(registers[4])
                data['Power'] = '0x{0:04x}'.format(registers[5])
                data['Diag_Alert'] = '0x{0:04x}'.format(registers[6])
                data['Current_OL'] = '0x{0:04x}'.format(registers[7])
                data['Current_UL'] = '0x{0:04x}'.format(registers[8])
                data['Voltage_OL'] = '0x{0:04x}'.format(registers[9])
                data['Voltage_UL'] = '0x{0:04x}'.format(registers[10])
                data['Temp_Limit'] = '0x{0:04x}'.format(registers[11])
                data['Power_Limit'] = '0x{0:04x}'.format(registers[12])
            case _:
                raise ValueError(f"{ps.part_name} Unknown part name for power sensor")
        return data

    def _set_power_sensor_conf(self, sensor_name, conf):
        ps = self.__find_power_sensor(sensor_name)
        if ps is None:
            raise ValueError(f'{sensor_name} does not exits')
        else:
            if ps._sensor is None:
                raise Exception(f'{sensor_name} sensor is not defined')
        match ps.part_name:
            case "INA226":
                print(conf)
                if len(conf) < 4:
                    raise ValueError(f"{conf} register value list is wrong")
            case "INA700" | "INA745A" | "INA745B":
                print(conf)
                if len(conf) < 5:
                    raise ValueError(f"{conf} register value list is wrong")
        return self.pmic.SetPowerSensorConf(ps._sensor, conf)

    def __find_voltage(self, voltage_name):
        temp_voltage = None
        for v in self.pmic.voltages:
            if v.name == voltage_name:
                temp_voltage = v
        return temp_voltage

    def _list_voltages(self):
        if len(self.pmic.voltages) == 0:
            raise ValueError(f"Voltage list is empty")
        data = []
        for v in self.pmic.voltages:
            data.append({
                v.name : {
                    'typical_volt' : self._scale(v.typical_volt, "voltage")
                }
            })
        return data

    def _enable_voltage(self, voltage_name):
        self.logger.info(f"EnableVoltage({voltage_name})")
        v = self.__find_voltage(voltage_name)
        if v is None:
            raise ValueError(f'{voltage_name} does not exits')
        else:
            if v._output is None:
                raise Exception(f'{voltage_name} regulator is not defined')
            else:
                self.pmic.EnableVoltage(v._output)

    def _disable_voltage(self, voltage_name):
        v = self.__find_voltage(voltage_name)
        if v is None:
            raise ValueError(f'{voltage_name} does not exits')
        else:
            if v._output is None:
                raise Exception(f'{voltage_name} regulator is not defined')
            else:
                self.pmic.DisableVoltage(v._output)

    def _get_regulator(self, voltage_name):
        v = self.__find_voltage(voltage_name)
        if v is None:
            raise ValueError(f'{voltage_name} does not exits')
        else:
            if v._output is None:
                raise Exception(f'{voltage_name} regulator is not defined')
            else:
                regulator_vals = self.pmic.GetRegulator(v._output)
                data = {
                    "vin" : self._scale(regulator_vals[0], "voltage"),
                    "iin" : self._scale(regulator_vals[1], "current"),
                    "vout" : self._scale(regulator_vals[2], "voltage"),
                    "iout" : self._scale(regulator_vals[3], "current"),
                    "temp" : self._scale(regulator_vals[4], "temperature"),
                    "pout" : self._scale(regulator_vals[5], "power"),
                    "pin" : self._scale(regulator_vals[6], "power"),
                }
        return data

    def _get_voltage(self, voltage_name):
        v = self.__find_voltage(voltage_name)
        if v is None:
            raise ValueError(f'{voltage_name} does not exits')
        else:
            if v._output is None:
                raise Exception(f'{voltage_name} regulator is not defined')
            else:
                data = {"Voltage" : self._scale(self.pmic.GetVoltage(v._output), "voltage")}
        return data

    def _set_voltage(self, voltage_name, new_value):
        v = self.__find_voltage(voltage_name)
        if v is None:
            raise ValueError(f'{voltage_name} does not exits')
        else:
            if new_value < self._scale(v.minimum_volt, "voltage"):
                raise ValueError(f'asked value({new_value}) below minimum({self._scale(v.minimum_volt, "voltage")}) for {v.name}')
            elif new_value > self._scale(v.maximum_volt, "voltage"):
                raise ValueError(f'asked value({new_value}) above maximum({self._scale(v.maximum_volt, "voltage")}) for {v.name}')
            else:
                if v._output is None:
                    raise Exception(f'{voltage_name} regulator is not defined')
                else:
                    self.pmic.SetVoltage(v._output, self._unscale(new_value, "voltage"))

    def _set_boot_voltage(self, voltage_name, boot_value):
        self.logger.info(f"SetBootVoltage({voltage_name}, {boot_value})")
        v = self.__find_voltage(voltage_name)
        if v is None:
            raise ValueError(f'{voltage_name} does not exits')
        else:
            if boot_value < self._scale(v.minimum_volt, "voltage"):
                raise ValueError(f'asked value({boot_value}) below minimum({self._scale(v.minimum_volt, "voltage")}) for {v.name}')
            elif boot_value > self._scale(v.maximum_volt, "voltage"):
                raise ValueError(f'asked value({boot_value}) above maximum({self._scale(v.maximum_volt, "voltage")}) for {v.name}')
            else:
                if v._output is None:
                    raise Exception(f'{voltage_name} regulator is not defined')

        directory = os.path.dirname(RAFT_DIR + '.raft/')
        if directory and not os.path.exists(directory):
            os.makedirs(directory)
        bootvoltage_path = os.path.join(directory, 'bootvoltage')

        try:
            with dbm.dumb.open(bootvoltage_path, 'c') as db:
                db[voltage_name] = "{:.3f}".format(round(self._unscale(boot_value, "voltage"), 3))
        except Exception as err:
            raise Exception(f"Unknown error: {err}")
        finally:
            db.sync()
            db.close()

    def _restore_voltage(self, voltage_name):
        self.logger.info(f"RestoreVoltage({voltage_name})")
        v = self.__find_voltage(voltage_name)
        if v is None:
            raise ValueError(f'{voltage_name} does not exits')
        else:
            self._set_voltage(voltage_name, self._scale(v.typical_volt, "voltage"))
            directory = os.path.dirname(RAFT_DIR + '.raft/')
            bootvoltage_path = os.path.join(directory, 'bootvoltage')
            try:
                with dbm.dumb.open(bootvoltage_path, 'w') as db:
                    del db[voltage_name]
            except KeyError:
                raise Exception(f"{voltage_name} voltage is not in bootvoltage list")
            finally:
                db.sync()
                db.close()

    def __find_temperature(self, temp_name):
        temp = None
        for v in self.temps:
            if v['Name'] == temp_name:
                temp = v
        return temp

    def _list_temperatures(self):
        self.logger.info(f"ListTemperatures()")
        data = []
        if len(self.temps) == 0:
            raise ValueError(f"Temperature list is empty")
        for s in self.temps:
            data.append(s['Name'])
        return data

    def _get_temperature(self, temp_name):
        self.logger.info(f"GetTemperature()")
        data = {}
        v = self.__find_temperature(temp_name)
        if v is None:
            raise ValueError(f'{temp_name} does not exits')
        else:
            temp, minimum, max_max, min_min = v['Sensor'].ReadSysmonTemperatures()
            data['TEMP'] = self._scale(temp, "temperature")
            data['MIN'] = self._scale(minimum, "temperature")
            data['MAX_MAX'] = self._scale(max_max, "temperature")
            data['MIN_MIN'] = self._scale(min_min, "temperature")
        return data

    def __del__(self):
        self.logger.info("Inside PM Destructor")


# Copyright (C) 2021-2022 Xilinx, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

__author__ = "Hugh Maguire"
__copyright__ = "Copyright 2021, Xilinx"

import Pyro4
import logging
import base64

logging.basicConfig(level=logging.ERROR)

class I2C_Client(object):
    I2C = None

    def __init__(self):
        logging.info("Inside CCF_Client Constructor")
        pass

    def SetIpAndPort(self, ipaddr, port):
        """
        API to inform I2C Client the IP address and port number of I2C Server.

        :param ipaddr: IP Address string
        :param port: Port number string
        :return: None

        """
        logging.debug("ipaddr = " + str(ipaddr))
        logging.debug("port = " + str(port))
        uri = "PYRO:I2C@" + str(ipaddr) + ":" + str(port)
        logging.debug("uri = " + uri)
        self.I2C = Pyro4.Proxy(uri)
        pass

    def XI2c_Initialize(self, device):
        """
        I2C driver one time initialisation.

        :param device_id: id of the opened device
        :return: ret_code
        """
        logging.debug(
            "XDfeCcf_Initialize(" + str(device) +  ")"
        )
        ret = self.I2C.XI2c_Initialize(device)
        logging.debug("The return value Init = " + str(ret))
        return ret

    def XI2c_ReadINA226Reg(self, addr, reg):
        """
        I2C driver read a register

        :param addr: address of INA226 device
        :param reg: register to be read
        :return: ret_code, value of register
        """
        ret, val = self.I2C.XI2c_ReadINA226Reg(addr, reg)
        return ret, val

    def XI2c_WriteINA226Reg(self, addr, reg, val):
        """
        I2C driver write a register

        :param addr: address of INA226 device
        :param reg: register to be written
        :param val: value to be written
        :return: ret_code
        """
        ret = self.I2C.XI2c_WriteINA226Reg(addr, reg, val)
        return ret

    def XI2c_Release(self):
        """
        I2C driver Release instance
        :return: none
        """
        self.I2C.XI2c_Release()

    def __del__(self):
        logging.info("Inside I2C Destructor")

i2c = I2C_Client()



# Copyright (C) 2021-2022 Xilinx, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause

__author__ = "Anish Kadamathikuttiyil Karthikeyan Pillai"
__copyright__ = "Copyright 2021, Xilinx"

import sys
RAFT_DIR = '/usr/share/raft/'
sys.path.append(RAFT_DIR + 'xserver/utils')
import logging
from utils import ffi, open_c_library
import logging

axi_memmap_handle = open_c_library(
    RAFT_DIR + "xserver/xcffi/drv_header/data_stream/data_transfer_no_dma/axi_memmap.h",
    "/usr/lib/libaximemmap.so.1",
)


class AXI_MEMMAP_C(object):
    def __init__(self):
        logging.info("Inside AXI_MEMMAP_C Constructor")
        pass

    def axi_read_words(self, address, num_words, network_order):
        """
        Read num_words from address.

        :param address: address to read from
        :param num_words: number of words to read
        :param network_order: Apply network byte ordering if network_order!=0
        :return: ret: whether success or failure
                 buf: the read data

        """
        buf = ffi.new("unsigned int[]", num_words)
        ret = axi_memmap_handle.axi_read_words(address, num_words, buf, network_order)
        buf = [buf[k] for k in range(num_words)]
        if ret == 0:
            logging.info("axi_read_words returned success")
        else:
            logging.info("axi_read_words returned failure ret = ", ret)
        return ret, buf

    def axi_write_words(self, address, num_words, buf, network_order):
        """
        Write num_words to address.

        :param address: address to write
        :param num_words: number of words to write
        :param buffer: data to write
        :param network_order: Apply network byte ordering if network_order!=0
        :return: ret: whether success or failure

        """
        buf_in = ffi.new("unsigned int[]", buf)
        ret = axi_memmap_handle.axi_write_words(
            address, num_words, buf_in, network_order
        )
        if ret == 0:
            logging.info("axi_write_words returned success")
        else:
            logging.info("axi_write_words returned failure ret = ", ret)
        return ret

    def __del__(self):
        logging.info("Inside AXI_MEMMAP_C Destructor")



# Copyright (C) 2021-2022 Xilinx, Inc.  All rights reserved.
# Copyright (C) 2022-2024 Advanced Micro Devices, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-3-Clause

__author__ = "Anish Kadamathikuttiyil Karthikeyan Pillai"
__copyright__ = "Copyright 2021, Xilinx"

import Pyro4
import logging
import base64
import json
import sys


class MIX_Client(object):
    MIX = None
    logger = None
    LogLevelsDict = {
        "DEBUG": 4,
        "INFO": 3,
        "WARNING": 2,
        "ERROR": 1,
        "CRITICAL": 0
    }

    def __init__(self):
        self.logger = self.GetLogger()
        self.logger.info("Inside Mixer Pyro Client Constructor")
        return

    @staticmethod
    def GetLogger():
        """
        Static method to get the logger for the class.
        Default loglevel is set inside this class

        :return: logger

        """
        log_level = logging.ERROR
        logging.basicConfig(format="%(levelname)s:%(message)s")
        logger = logging.getLogger(__name__)
        try:
            handler_set_check = getattr(logger, 'handler_set')
        except AttributeError:
            handler_set_check = False
        if not handler_set_check:
            logger.setLevel(log_level)
            handler = logging.StreamHandler(sys.stdout)
            logger.addHandler(handler)
            logger.handler_set = True
            logger.disabled = False
        return logger

    @staticmethod
    def extract_b64_encoded_string(DictName):
        b64_data = DictName['data']
        b64_bytes_data = b64_data.encode('ascii')
        msg_bytes_data = base64.b64decode(b64_bytes_data)
        data = msg_bytes_data.decode('ascii')
        data = bytes(data, 'utf-8')
        return data

    def SetIpAndPort(self, ipaddr, port):
        """
        API to inform MIXER Client the IP address and port number of MIXER Server.

        :param ipaddr: IP Address string
        :param port: Port number string
        :return: None

        """    
        uri = f"PYRO:MIX@{ipaddr}:{port}"
        logging.debug(f"SetIpAndPort({ipaddr}, {port})\n uri = {uri}")   
        self.MIX = Pyro4.Proxy(uri)
        pass

    def GetPythonLogLevels(self):
        """
        Return the logging levels supported by logging library in python

        :param : None
        :return: Dictionary showing the log levels supported by logging library
        """
        self.logger.debug("GetPythonLogLevels()")
        PythonLogLevel = self.MIX.GetPythonLogLevels()
        self.logger.debug(f"PythonLogLevel = {json.dumps(PythonLogLevel, indent=2)}")
        return PythonLogLevel

    def SetServerLogLevel(self, PythonLogLevel):
        """
        Set the python log level to the given level

        :param : Log level to set
        :return: None
        """
        self.logger.debug(f"SetServerLogLevel({PythonLogLevel})")
        self.MIX.SetServerLogLevel(PythonLogLevel)
        return

    def SetClientLogLevel(self, PythonLogLevel):
        """
        Set the python log level to the given level

        :param : Log level to set
        :return: None
        """
        if PythonLogLevel == self.LogLevelsDict["DEBUG"]:
            self.logger.setLevel(logging.DEBUG)
        elif PythonLogLevel == self.LogLevelsDict["INFO"]:
            self.logger.setLevel(logging.INFO)
        elif PythonLogLevel == self.LogLevelsDict["WARNING"]:
            self.logger.setLevel(logging.WARNING)
        elif PythonLogLevel == self.LogLevelsDict["ERROR"]:
            self.logger.setLevel(logging.ERROR)
        else:
            self.logger.setLevel(logging.CRITICAL)
        return

    def SetMetalLogLevel(self, MetalLogLevel):
        """
        Set the metal log level to the given level

        :param : Log level to set
        :return: None
        """
        self.logger.debug(f"SetMetalLogLevel({MetalLogLevel})")
        self.MIX.SetMetalLogLevel(MetalLogLevel)
        return

    def GetMixMacro(self):
        """
        Return Dictionary with all MIX macros in the mix header file

        :param : None
        :return: Dictionary with all MIX macros in the mix header file
        """
        self.logger.debug("GetMixMacro()")
        mix_macro = self.MIX.GetMixMacro()
        self.logger.debug(f"mix_macro = {json.dumps(mix_macro, indent=2)}")
        return mix_macro

    # Get enum API
    def GetEnum_XDfeMix_StateId(self):
        """
        Return Dictionary equivalent of enum XDfeMix_StateId

        :param : None
        :return: Dictionary equivalent of enum XDfeMix_StateId
        """
        self.logger.debug("GetEnum_XDfeMix_StateId()")
        XDfeMix_StateId = self.MIX.GetEnum_XDfeMix_StateId()
        self.logger.debug(f"XDfeMix_StateId = {json.dumps(XDfeMix_StateId, indent=2)}")
        return XDfeMix_StateId

    def GetEnum_metal_log_level(self):
        """
        Return Dictionary equivalent of enum metal_log_level

        :param : None
        :return: Dictionary equivalent of enum metal_log_level
        """
        self.logger.debug("GetEnum_metal_log_level()")
        metal_log_level = self.MIX.GetEnum_metal_log_level()
        self.logger.debug(f"metal_log_level = {json.dumps(metal_log_level, indent=2)}")
        return metal_log_level

    # Get structure  API
    def GetStruct_XDfeMix_Version(self):
        """
        Return Dictionary equivalent of structure XDfeMix_Version

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_Version
        """
        self.logger.debug("GetStruct_XDfeMix_Version()")
        XDfeMix_Version = self.MIX.GetStruct_XDfeMix_Version()
        return XDfeMix_Version

    def GetStruct_XDfeMix_Trigger(self):
        """
        Return Dictionary equivalent of structure XDfeMix_Trigger

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_Trigger
        """
        self.logger.debug("GetStruct_XDfeMix_Trigger()")
        XDfeMix_Trigger = self.MIX.GetStruct_XDfeMix_Trigger()
        return XDfeMix_Trigger

    def GetStruct_XDfeMix_TriggerCfg(self):
        """
        Return Dictionary equivalent of structure XDfeMix_TriggerCfg

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_TriggerCfg
        """
        self.logger.debug("GetStruct_XDfeMix_TriggerCfg()")
        XDfeMix_TriggerCfg = self.MIX.GetStruct_XDfeMix_TriggerCfg()
        return XDfeMix_TriggerCfg

    def GetStruct_XDfeMix_CCSequence(self):
        """
        Return Dictionary equivalent of structure XDfeMix_CCSequence

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_CCSequence
        """
        self.logger.debug("GetStruct_XDfeMix_CCSequence()")
        XDfeMix_CCSequence = self.MIX.GetStruct_XDfeMix_CCSequence()
        return XDfeMix_CCSequence

    def GetStruct_XDfeMix_ModelParameters(self):
        """
        Return Dictionary equivalent of structure XDfeMix_ModelParameters

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_ModelParameters
        """
        self.logger.debug("GetStruct_XDfeMix_ModelParameters()")
        XDfeMix_ModelParameters = self.MIX.GetStruct_XDfeMix_ModelParameters()
        return XDfeMix_ModelParameters

    def GetStruct_XDfeMix_Cfg(self):
        """
        Return Dictionary equivalent of structure XDfeMix_Cfg

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_Cfg
        """
        self.logger.debug("GetStruct_XDfeMix_Cfg()")
        XDfeMix_Cfg = self.MIX.GetStruct_XDfeMix_Cfg()
        return XDfeMix_Cfg

    def GetStruct_XDfeMix_Init(self):
        """
        Return Dictionary equivalent of structure XDfeMix_Init

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_Init
        """
        self.logger.debug("GetStruct_XDfeMix_Init()")
        XDfeMix_Init = self.MIX.GetStruct_XDfeMix_Init()
        return XDfeMix_Init

    def GetStruct_XDfeMix_PhaseOffset(self):
        """
        Return Dictionary equivalent of structure XDfeMix_PhaseOffset

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_PhaseOffset
        """
        self.logger.debug("GetStruct_XDfeMix_PhaseOffset()")
        XDfeMix_PhaseOffset = self.MIX.GetStruct_XDfeMix_PhaseOffset()
        return XDfeMix_PhaseOffset

    def GetStruct_XDfeMix_Frequency(self):
        """
        Return Dictionary equivalent of structure XDfeMix_Frequency

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_Frequency
        """
        self.logger.debug("GetStruct_XDfeMix_Frequency()")
        XDfeMix_Frequency = self.MIX.GetStruct_XDfeMix_Frequency()
        return XDfeMix_Frequency

    def GetStruct_XDfeMix_Phase(self):
        """
        Return Dictionary equivalent of structure XDfeMix_Phase

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_Phase
        """
        self.logger.debug("GetStruct_XDfeMix_Phase()")
        XDfeMix_Phase = self.MIX.GetStruct_XDfeMix_Phase()
        return XDfeMix_Phase

    def GetStruct_XDfeMix_NCO(self):
        """
        Return Dictionary equivalent of structure XDfeMix_NCO

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_NCO
        """
        self.logger.debug("GetStruct_XDfeMix_NCO()")
        XDfeMix_NCO = self.MIX.GetStruct_XDfeMix_NCO()
        return XDfeMix_NCO

    def GetStruct_XDfeMix_InternalDUCDDCCfg(self):
        """
        Return Dictionary equivalent of structure XDfeMix_InternalDUCDDCCfg

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_InternalDUCDDCCfg
        """
        self.logger.debug("GetStruct_XDfeMix_InternalDUCDDCCfg()")
        XDfeMix_InternalDUCDDCCfg = self.MIX.GetStruct_XDfeMix_InternalDUCDDCCfg()
        return XDfeMix_InternalDUCDDCCfg

    def GetStruct_XDfeMix_DUCDDCCfg(self):
        """
        Return Dictionary equivalent of structure XDfeMix_DUCDDCCfg

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_DUCDDCCfg
        """
        self.logger.debug("GetStruct_XDfeMix_DUCDDCCfg()")
        XDfeMix_DUCDDCCfg = self.MIX.GetStruct_XDfeMix_DUCDDCCfg()
        return XDfeMix_DUCDDCCfg

    def GetStruct_XDfeMix_AuxiliaryCfg(self):
        """
        Return Dictionary equivalent of structure XDfeMix_AuxiliaryCfg

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_AuxiliaryCfg
        """
        self.logger.debug("GetStruct_XDfeMix_AuxiliaryCfg()")
        XDfeMix_AuxiliaryCfg = self.MIX.GetStruct_XDfeMix_AuxiliaryCfg()
        return XDfeMix_AuxiliaryCfg

    def GetStruct_XDfeMix_CarrierCfg(self):
        """
        Return Dictionary equivalent of structure XDfeMix_CarrierCfg

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_CarrierCfg
        """
        self.logger.debug("GetStruct_XDfeMix_CarrierCfg()")
        XDfeMix_CarrierCfg = self.MIX.GetStruct_XDfeMix_CarrierCfg()
        return XDfeMix_CarrierCfg

    def GetStruct_XDfeMix_AntennaCfg(self):
        """
        Return Dictionary equivalent of structure XDfeMix_AntennaCfg

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_AntennaCfg
        """
        self.logger.debug("GetStruct_XDfeMix_AntennaCfg()")
        XDfeMix_AntennaCfg = self.MIX.GetStruct_XDfeMix_AntennaCfg()
        return XDfeMix_AntennaCfg

    def GetStruct_XDfeMix_CCCfg(self):
        """
        Return Dictionary equivalent of structure XDfeMix_CCCfg

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_CCCfg
        """
        self.logger.debug("GetStruct_XDfeMix_CCCfg()")
        XDfeMix_CCCfg = self.MIX.GetStruct_XDfeMix_CCCfg()
        return XDfeMix_CCCfg

    def GetStruct_XDfeMix_DUCDDCStatus(self):
        """
        Return Dictionary equivalent of structure XDfeMix_DUCDDCStatus

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_DUCDDCStatus
        """
        self.logger.debug("GetStruct_XDfeMix_DUCDDCStatus()")
        XDfeMix_DUCDDCStatus = self.MIX.GetStruct_XDfeMix_DUCDDCStatus()
        return XDfeMix_DUCDDCStatus

    def GetStruct_XDfeMix_MixerStatus(self):
        """
        Return Dictionary equivalent of structure XDfeMix_MixerStatus

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_MixerStatus
        """
        self.logger.debug("GetStruct_XDfeMix_MixerStatus()")
        XDfeMix_MixerStatus = self.MIX.GetStruct_XDfeMix_MixerStatus()
        return XDfeMix_MixerStatus

    def GetStruct_XDfeMix_Status(self):
        """
        Return Dictionary equivalent of structure XDfeMix_Status

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_Status
        """
        self.logger.debug("GetStruct_XDfeMix_Status()")
        XDfeMix_Status = self.MIX.GetStruct_XDfeMix_Status()
        return XDfeMix_Status

    def GetStruct_XDfeMix_InterruptMask(self):
        """
        Return Dictionary equivalent of structure XDfeMix_InterruptMask

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_InterruptMask
        """
        self.logger.debug("GetStruct_XDfeMix_InterruptMask()")
        XDfeMix_InterruptMask = self.MIX.GetStruct_XDfeMix_InterruptMask()
        return XDfeMix_InterruptMask

    def GetStruct_XDfeMix_Config(self):
        """
        Return Dictionary equivalent of structure XDfeMix_Config

        :param : None
        :return: Dictionary equivalent of structure XDfeMix_Config
        """
        self.logger.debug("GetStruct_XDfeMix_Config()")
        XDfeMix_Config = self.MIX.GetStruct_XDfeMix_Config()
        return XDfeMix_Config

    def GetStruct_XDfeMix(self):
        """
        Return Dictionary equivalent of structure XDfeMix

        :param : None
        :return: Dictionary equivalent of structure XDfeMix
        """
        self.logger.debug("GetStruct_XDfeMix()")
        XDfeMix = self.MIX.GetStruct_XDfeMix()
        return XDfeMix

    # System initialization API
    def XDfeMix_InstanceInit(self, DeviceNodeName):
        """
        API initialises one instance of a Mixer driver.
        Traverses "/sys/bus/platform/device" directory (in Linux), to find registered
        XDfeMix device with the name DeviceNodeName. The first available slot in
        the instances array XDfeMix_Mixer[] will be taken as a DeviceNodeName
        object. On success it moves the state machine to a Ready state, while on
        failure stays in a Not Ready state.

        :param DeviceNodeName: device node name.
        :return: ret - 0 on success, 1 on failure
                 device_id - integer handle to initialized instance
                 DeviceNodeNameRet - device node name returned from driver
                                     which will be same as the passed value

        """
        self.logger.info(f"XDfeMix_InstanceInit({DeviceNodeName})")
        DeviceNodeName = bytes(str(DeviceNodeName), 'utf-8')
        ret, device_id, DeviceNodeName = self.MIX.XDfeMix_InstanceInit(DeviceNodeName)
        DeviceNodeName = self.extract_b64_encoded_string(DeviceNodeName)
        self.logger.info(f"ret = {ret}, device_id = {device_id}, DeviceNodeName = {DeviceNodeName}")
        return ret, device_id, DeviceNodeName

    def XDfeMix_InstanceClose(self, device_id):
        """
        API closes the instances of a Mixer driver and moves the state machine to
        a Not Ready state.

        :param device_id: id of the opened device.
        :return: None
        """
        self.logger.info(f"XDfeMix_InstanceClose({device_id})")
        self.MIX.XDfeMix_InstanceClose(device_id)
        return

    # Register access API
    def XDfeMix_WriteReg(self, device_id, addr_offset, data):
        """
        Writes a value to register in a Mixer instance.

        :param device_id: id of the opened device.
        :param AddrOffset: is address offset relative to instance base address.
        :param Data: is value to be written.
        :return: None
        """
        self.logger.debug(f"XDfeMix_WriteReg({device_id}, {hex(addr_offset)}, {hex(data)})")
        self.MIX.XDfeMix_WriteReg(device_id, addr_offset, data)
        return

    def XDfeMix_ReadReg(self, device_id, addr_offset):
        """
        Reads a value from the register from a Mixer instance.

        :param device_id: id of the opened device.
        :param AddrOffset: is address offset relative to instance base address.
        :return: regval: Register value.
        """
        self.logger.debug(f"XDfeMix_ReadReg({device_id}, {hex(addr_offset)})")
        regval = self.MIX.XDfeMix_ReadReg(device_id, addr_offset)
        self.logger.debug(f"regval = {regval}")
        return regval

    # DFE Mixer component initialization API
    def XDfeMix_Reset(self, device_id):
        """
        Resets and put block into a reset state.

        :param device_id: id of the opened device.
        :return: None
        """
        self.logger.debug(f"XDfeMix_Reset({device_id})")
        self.MIX.XDfeMix_Reset(device_id)
        return

    def XDfeMix_Configure(self, device_id, Cfg):
        """
        Reads configuration from device tree/xparameters.h and IP registers.
        Removes S/W reset and moves the state machine to a Configured state.

        :param device_id: id of the opened device.
        :param Cfg: configuration data container.
        :return: Cfg: configuration data container.
        """
        self.logger.debug(f"XDfeMix_Configure({device_id}, {json.dumps(Cfg, indent=2)})")
        Cfg = self.MIX.XDfeMix_Configure(device_id, Cfg)
        self.logger.debug(f"Cfg = {json.dumps(Cfg, indent=2)}")
        return Cfg

    def XDfeMix_Initialize(self, device_id, Init):
        """
        DFE Mixer driver one time initialisation and moves the state machine to
        an Initialised state.

        :param device_id: id of the opened device.
        :param Init: initialisation data container.
        :return: Init: initialisation data container.
        """
        self.logger.debug(f"XDfeMix_Initialize({device_id}, {json.dumps(Init, indent=2)})")
        Init = self.MIX.XDfeMix_Initialize(device_id, Init)
        self.logger.debug(f"Init = {json.dumps(Init, indent=2)}")
        return Init

    def XDfeMix_Activate(self, device_id, EnableLowPower):
        """
        Activates Mixer and moves the state machine to an Activated state.

        :param device_id: id of the opened device.
        :param EnableLowPower: flag indicating low power.
        :return: None
        """
        self.logger.debug(f"XDfeMix_Activate({device_id}, {EnableLowPower})")
        self.MIX.XDfeMix_Activate(device_id, EnableLowPower)
        return

    def XDfeMix_Deactivate(self, device_id):
        """
        Deactivates Mixer and moves the state machine to Initialised state.

        :param device_id: id of the opened device.
        :return: None
        """
        self.logger.debug(f"XDfeMix_Deactivate({device_id})")
        self.MIX.XDfeMix_Deactivate(device_id)
        return

    def XDfeMix_GetStateID(self, device_id):
        """
        Gets a state machine state id. The state id is returned as a string.
        The enum is mapped to a dictionary in python.

        :param device_id: id of the opened device.
        :return: StateId: State machine StateID as a string
        """
        self.logger.debug(f"XDfeMix_GetStateID({device_id})")
        StateId = self.MIX.XDfeMix_GetStateID(device_id)
        self.logger.debug(f"Return value StateId = {StateId}")
        return StateId

    def XDfeMix_GetCurrentCCCfg(self, device_id, CurrCCCfg):
        """
        Returns the current CC configuration. Not used slot ID in a sequence
        (Sequence.CCID[Index]) are represented as (-1), not the value in registers.

        :param device_id: id of the opened device.
        :param CurrCCCfg: CC configuration container.
        :return: CurrCCCfg: CC configuration container
        """
        self.logger.debug(f"XDfeMix_GetCurrentCCCfg({device_id}, {json.dumps(CurrCCCfg, indent=2)})")
        CurrCCCfg = self.MIX.XDfeMix_GetCurrentCCCfg(device_id, CurrCCCfg)
        self.logger.debug(f"CurrCCCfg = {json.dumps(CurrCCCfg, indent=2)}")
        return CurrCCCfg

    def XDfeMix_GetCurrentCCCfgSwitchable(self, device_id, CCCfgDownlink, CCCfgUplink):
        """
		Returns the current CC configuration for DL and UL in switchable mode.
		Not used slot ID in a sequence (Sequence.CCID[Index]) are represented
		as (-1), not the value in registers.

        :param device_id: id of the opened device.
		:param CCCfgDownlink: Downlink CC configuration container.
		:param CCCfgUplink: Uplink CC configuration container.
        :return: CCCfgDownlink, CCCfgUplink: CC configuration containers
        """
        self.logger.debug(f"XDfeMix_GetCurrentCCCfgSwitchable({device_id}, "
						  f"{json.dumps(CCCfgDownlink, indent=2)}, "
						  f"{json.dumps(CCCfgUplink, indent=2)})")
        CCCfgDownlink, CCCfgUplink = self.MIX.XDfeMix_GetCurrentCCCfgSwitchable(device_id, CCCfgDownlink, CCCfgUplink)
        self.logger.debug(f"CCCfgDownlink = {json.dumps(CCCfgDownlink, indent=2)}")
        self.logger.debug(f"CCCfgUplink = {json.dumps(CCCfgUplink, indent=2)}")
        return CCCfgDownlink, CCCfgUplink

    def XDfeMix_GetEmptyCCCfg(self, device_id):
        """
        Returns configuration structure CCCfg with CCCfg->Sequence.Length value set
        in XDfeMix_Configure(), array CCCfg->Sequence.CCID[] members are set to not
        used value (-1) and the other CCCfg members are set to 0.

        :param device_id: id of the opened device.
        :return: CCCfg: CC configuration container
        """
        self.logger.debug(f"XDfeMix_GetEmptyCCCfg({device_id})")
        CCCfg = self.MIX.XDfeMix_GetEmptyCCCfg(device_id)
        self.logger.debug(f"CCCfg = {json.dumps(CCCfg, indent=2)}")
        return CCCfg

    def XDfeMix_GetCarrierCfgAndNCO(self, device_id, CCCfg, CCID):
        """
        Returns the current CC sequence bitmap, CCID carrier configuration and
        NCO configuration.

        :param device_id: id of the opened device.
        :param CCCfg: component carrier (CC) configuration container.
        :param CCID: Channel ID.
        :return: CCSeqBitmap: CC slot position container.
                 CarrierCfg: CC configuration container.
                 NCO: NCO configuration container.
        """
        self.logger.debug(f"XDfeMix_GetCarrierCfgAndNCO({device_id}, {json.dumps(CCCfg, indent=2)}, {CCID})")
        CCSeqBitmap, CarrierCfg, NCO = self.MIX.XDfeMix_GetCarrierCfgAndNCO(device_id, CCCfg, CCID)
        self.logger.debug(f"Return value = {CCSeqBitmap}, {json.dumps(CarrierCfg, indent=2)},"
                          f" {json.dumps(NCO, indent=2)}")
        return CCSeqBitmap, CarrierCfg, NCO

    def XDfeMix_SetAntennaCfgInCCCfg(self, device_id, CCCfg, AntennaCfg):
        """
        Set antenna configuration in CC configuration container.

        :param device_id: id of the opened device.
        :param CCCfg: component carrier (CC) configuration container.
        :param AntennaCfg: Array of all antenna configurations.
        :return: CCCfg: component carrier (CC) configuration container.
        """
        self.logger.debug(f"XDfeMix_SetAntennaCfgInCCCfg({device_id}, {json.dumps(CCCfg, indent=2)}, {json.dumps(AntennaCfg, indent=2)})")
        CCCfg = self.MIX.XDfeMix_SetAntennaCfgInCCCfg(device_id, CCCfg, AntennaCfg)
        self.logger.debug(f"Return value = {json.dumps(CCCfg, indent=2)}")
        return CCCfg

    def XDfeMix_AddCCtoCCCfg(self, device_id, CCCfg, CCID, CCSeqBitmap, CarrierCfg, NCO):
        """
        Adds specified CCID, with specified configuration, to a local CC
        configuration structure.
        If there is insufficient capacity for the new CC the function will return
        an error.
        Initiates CC update (enable CCUpdate trigger TUSER Single Shot).

        The returned CCCfg.Sequence is transleted as there is no explicit indication
        that SEQUENCE[i] is not used - 0 can define the slot as either used or
        not used. Sequence data that is returned in the CCIDSequence is not the same
        as what is written in the registers. The translation is:
            - CCIDSequence.CCID[i] = -1    - if [i] is unused slot
            - CCIDSequence.CCID[i] = CCID  - if [i] is used slot
            - a returned CCIDSequence->Length = length in register + 1

        :param device_id: id of the opened device.
        :param CCCfg: component carrier (CC) configuration container.
        :param CCID: Channel ID.
        :param CCSeqBitmap: CC slot position container.
        :param CarrierCfg: CC configuration container.
        :param NCO: NCO configuration container.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
                 CCCfg: component carrier (CC) configuration container.
        """
        self.logger.debug(f"XDfeMix_AddCCtoCCCfg({device_id}, {json.dumps(CCCfg, indent=2)}, "
                          f"{CCID}, {CCSeqBitmap}, {json.dumps(CarrierCfg, indent=2)}, "
                          f"{json.dumps(NCO, indent=2)})")
        ret, CCCfg = self.MIX.XDfeMix_AddCCtoCCCfg(device_id, CCCfg, CCID, CCSeqBitmap,
                                                   CarrierCfg, NCO)
        self.logger.debug(f"Return value = {ret}, {json.dumps(CCCfg, indent=2)}")
        return ret, CCCfg

    def XDfeMix_RemoveCCfromCCCfg(self, device_id, CCCfg, CCID):
        """
        Removes specified CCID from a local CC configuration structure.

        :param device_id: id of the opened device.
        :param CCCfg: component carrier (CC) configuration container.
        :param CCID: Channel ID.
        :return: CCCfg: component carrier (CC) configuration container.
        """
        self.logger.debug(f"XDfeMix_RemoveCCfromCCCfg({device_id}, "
                          f"{json.dumps(CCCfg, indent=2)}, {CCID})")
        CCCfg = self.MIX.XDfeMix_RemoveCCfromCCCfg(device_id, CCCfg, CCID)
        self.logger.debug(f"Return value = {json.dumps(CCCfg, indent=2)}")
        return CCCfg

    def XDfeMix_AddAuxNCOtoCCCfg(self, device_id, CCCfg, AuxId, NCO, AuxCfg):
        """
		Adds specified auxiliary NCO, with specified configuration, to a local CCCfg.

        :param device_id: id of the opened device.
        :param CCCfg: component carrier (CC) configuration container.
        :param AuxId: Channel ID.
        :param NCO: Auxiliary NCO configuration container.
        :param AuxCfg: Auxiliary NCO configuration container.
        :return: CCCfg: component carrier (CC) configuration container.
        """
        self.logger.debug(f"XDfeMix_AddAuxNCOtoCCCfg({device_id}, {json.dumps(CCCfg, indent=2)}, "
                          f"{AuxId}, {json.dumps(AuxCfg, indent=2)})")
        CCCfg = self.MIX.XDfeMix_AddAuxNCOtoCCCfg(device_id, CCCfg, AuxId, NCO, AuxCfg)
        self.logger.debug(f"Return value = {json.dumps(CCCfg, indent=2)}")
        return CCCfg

    def XDfeMix_RemoveAuxNCOfromCCCfg(self, device_id, CCCfg, AuxId):
        """
		Disables specified auxiliary NCO configuration structure.

        :param device_id: id of the opened device.
        :param CCCfg: component carrier (CC) configuration container.
        :param AuxId: Auxiliary NCO ID to be disabled, range [0-3].
        :return: CCCfg: component carrier (CC) configuration container.
        """
        self.logger.debug(f"XDfeMix_RemoveAuxNCOfromCCCfg({device_id}, "
                          f"{json.dumps(CCCfg, indent=2)}, {AuxId})")
        CCCfg = self.MIX.XDfeMix_RemoveAuxNCOfromCCCfg(device_id, CCCfg, AuxId)
        self.logger.debug(f"Return value = {json.dumps(CCCfg, indent=2)}")
        return CCCfg

    def XDfeMix_UpdateCCinCCCfg(self, device_id, CCCfg, CCID, CarrierCfg):
        """
        Updates specified CCID, with specified configuration to a local CC
        configuration structure.
        If there is insufficient capacity for the new CC the function will return
        an error.

        :param device_id: id of the opened device.
        :param CCCfg: component carrier (CC) configuration container.
        :param CCID: Channel ID.
        :param CarrierCfg: CC configuration container.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
                 CarrierCfg: is a CC configuration container.
        """
        self.logger.debug(f"XDfeMix_UpdateCCinCCCfg({device_id}, {json.dumps(CCCfg, indent=2)}, "
                          f"{CCID}, {json.dumps(CarrierCfg, indent=2)}")
        ret, CCCfg = self.MIX.XDfeMix_UpdateCCinCCCfg(device_id, CCCfg, CCID, CarrierCfg)
        self.logger.debug(f"Return value CCCfg = {json.dumps(CCCfg, indent=2)}")
        return ret, CCCfg

    def XDfeMix_SetNextCCCfgAndTrigger(self, device_id, CCCfg):
        """
        Writes local CC configuration to the shadow (NEXT) registers and triggers
        copying from shadow to operational registers.

        :param device_id: id of the opened device.
        :param CCCfg: component carrier (CC) configuration container.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
                 CCCfg: component carrier (CC) configuration container.
        """
        self.logger.debug(f"XDfeMix_SetNextCCCfgAndTrigger({device_id}, "
                          f"{json.dumps(CCCfg, indent=2)})")
        ret, CCCfg = self.MIX.XDfeMix_SetNextCCCfgAndTrigger(device_id, CCCfg)
        self.logger.debug(f"ret = {ret}, CCCfg = {json.dumps(CCCfg, indent=2)}")
        return ret, CCCfg

    def XDfeMix_SetNextCCCfgAndTriggerSwitchable(self, device_id, CCCfgDownlink, CCCfgUplink):
        """
		Writes local CC configuration to the shadow (NEXT) registers and triggers
		copying from shadow to operational (CURRENT) registers.

        :param device_id: id of the opened device.
        :param CCCfgDownlink: Downlink CC configuration container.
        :param CCCfgUplink Uplink CC configuration container.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
                 CCCfgDownlink: Downlink component carrier (CC) configuration container.
                 CCCfgUplink: Uplink component carrier (CC) configuration container.
        """
        self.logger.debug(f"XDfeMix_SetNextCCCfgAndTriggerSwitchable({device_id}, "
                          f"{json.dumps(CCCfgDownlink, indent=2)}, "
                          f"{json.dumps(CCCfgDownlink, indent=2)})")
        ret, CCCfgDownlink, CCCfgUplink = self.MIX.XDfeMix_SetNextCCCfgAndTriggerSwitchable(device_id, CCCfgDownlink, CCCfgUplink)
        self.logger.debug(f"ret = {ret}, CCCfgDownlink = {json.dumps(CCCfgDownlink, indent=2)}, , CCCfgUplink = {json.dumps(CCCfgUplink, indent=2)}")
        return ret, CCCfgDownlink, CCCfgUplink

    def XDfeMix_AddCC(self, device_id, CCID, BitSequence, CarrierCfg, NCO):
        """
        Adds specified CCID, with specified configuration.
        If there is insufficient capacity for the new CC the function will return
        an error.
        Initiates CC update (enable CCUpdate trigger TUSER Single Shot).
        Note: Clear event status with XDfeMix_ClearEventStatus() before running this API.

        :param device_id: id of the opened device.
        :param CCID: is a Channel ID.
        :param BitSequence: up to 16 defined slots into which a CC can be
                            allocated. The number of slots can be from 1 to 16 depending on
                            system initialization. The number of slots is defined by the
                            "sequence length" parameter which is provided during initialization.
                            The Bit offset within the CCSeqBitmap indicates the equivalent
                            Slot number to allocate. e.g. 0x0003  means the caller wants the
                            passed component carrier (CC) to be allocated to slots 0 and 1.
        :param CarrierCfg: is a CC configuration container.
        :param NCO: NCO configuration container.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
                 CarrierCfg: is a CC configuration container.
        """
        self.logger.debug(f"XDfeMix_AddCC({device_id}, {CCID}, {BitSequence}, "
                          f"{json.dumps(CarrierCfg, indent=2)}, "
                          f"{json.dumps(NCO, indent=2)})")
        ret, CarrierCfg = self.MIX.XDfeMix_AddCC(device_id, CCID, BitSequence, CarrierCfg, NCO)
        self.logger.debug(f"ret = {ret}, CarrerCfg = {json.dumps(CarrierCfg, indent=2)})")
        return ret, CarrierCfg

    def XDfeMix_RemoveCC(self, device_id, CCID):
        """
        Removes specified CCID.
        Initiates CC update (enable CCUpdate trigger TUSER Single Shot).
        Note: Clear event status with XDfeMix_ClearEventStatus() before running this API.

        :param device_id: id of the opened device.
        :param CCID: is a Channel ID.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
        """
        self.logger.debug(f"XDfeMix_RemoveCC({device_id}, {CCID})")
        ret = self.MIX.XDfeMix_RemoveCC(device_id, CCID)
        self.logger.debug(f"ret = {ret}")
        return ret

    def XDfeMix_MoveCC(self, device_id, CCID, Rate, FromNCO, ToNCO):
        """
        Moves specified CCID from one NCO to another aligning phase to make it
        transparent.
        Initiates CC update (enable CCUpdate trigger TUSER Single Shot).
        Note: Clear event status with XDfeMix_ClearEventStatus() before running this API.

        :param device_id: id of the opened device.
        :param CCID: is a Channel ID.
        :param Rate: is a NCO rate value.
        :param FromNCO: is a NCO value moving from.
        :param ToNCO: is a NCO value moving to.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
        """
        self.logger.debug(f"XDfeMix_MoveCC({device_id}, {CCID}, {Rate}, {FromNCO}, {ToNCO})")
        ret = self.MIX.XDfeMix_MoveCC(device_id, CCID, Rate, FromNCO, ToNCO)
        self.logger.debug(f"ret = {ret}")
        return ret

    def XDfeMix_UpdateCC(self, device_id, CCID, CarrierCfg):
        """
        Updates specified CCID, with a configuration defined in CarrierCfg
        structure.
        If there is insufficient capacity for the new CC the function will return
        an error.
        Note: Clear event status with XDfeMix_ClearEventStatus() before running this API.

        :param device_id: id of the opened device.
        :param CCID: is a Channel ID.
        :param CarrierCfg: is a CC configuration container.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
        """
        self.logger.debug(f"XDfeMix_UpdateCC({device_id}, {CCID}, "
                          f"{json.dumps(CarrierCfg, indent=2)})")
        ret = self.MIX.XDfeMix_UpdateCC(device_id, CCID, CarrierCfg)
        self.logger.debug(f"ret = {ret}")
        return ret

    def XDfeMix_SetAntennaGain(self, device_id, AntennaId, AntennaGain):
        """
        Sets antenna gain. Initiate CC update (enable CCUpdate trigger TUSER
        Single Shot).
        Note: Clear event status with XDfeMix_ClearEventStatus() before running this API.

        :param device_id: id of the opened device.
        :param AntennaId: is an antenna ID.
        :param AntennaGain: is an antenna gain.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
        """
        self.logger.debug(f"XDfeMix_SetAntennaGain({device_id}, {AntennaId}, {AntennaGain})")
        ret = self.MIX.XDfeMix_SetAntennaGain(device_id, AntennaId, AntennaGain)
        self.logger.debug(f"ret = {ret}")
        return ret

    def XDfeMix_UpdateAntennaCfg(self, device_id, AntennaCfg):
        """
        Updates antenna configuration to all antennas.
        Note: Clear event status with XDfeMix_ClearEventStatus() before running this API.

        :param device_id: id of the opened device.
        :param AntennaCfg: Array of all antenna configurations.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
        """
        self.logger.debug(f"XDfeMix_UpdateAntennaCfg({device_id}, {json.dumps(AntennaCfg, indent=2)})")
        ret = self.MIX.XDfeMix_UpdateAntennaCfg(device_id, AntennaCfg)
        self.logger.debug(f"ret = {ret}")
        return ret

    def XDfeMix_GetTriggersCfg(self, device_id):
        """
        Returns current trigger configuration.

        :param device_id: id of the opened device.
        :return: TriggerCfg: trigger configuration container
        """
        self.logger.debug(f"XDfeMix_GetTriggersCfg({device_id})")
        TriggerCfg = self.MIX.XDfeMix_GetTriggersCfg(device_id)
        self.logger.debug(f"TriggerCfg = {json.dumps(TriggerCfg, indent=2)}")
        return TriggerCfg

    def XDfeMix_SetTriggersCfg(self, device_id, TriggerCfg):
        """
        Sets trigger configuration.

        :param device_id: id of the opened device.
        :param TriggerCfg: trigger configuration container.
        :return: TriggerCfg: trigger configuration container.
        """
        self.logger.debug(f"XDfeMix_SetTriggersCfg({device_id}, {json.dumps(TriggerCfg, indent=2)})")
        TriggerCfg = self.MIX.XDfeMix_SetTriggersCfg(device_id, TriggerCfg)
        self.logger.debug(f"The return value TriggerCfg = {json.dumps(TriggerCfg, indent=2)}")
        return TriggerCfg

    def XDfeMix_GetDUCDDCStatus(self, device_id):
        """
        Gets DUC/DDC overflow status.

        :param device_id: id of the opened device.
        :return: DUCDDCStatus: DUC/DDC status container.
        """
        self.logger.debug(f"XDfeMix_GetDUCDDCStatus({device_id})")
        DUCDDCStatus = self.MIX.XDfeMix_GetDUCDDCStatus(device_id)
        self.logger.debug(f"The return value DUCDDCStatus = {json.dumps(DUCDDCStatus, indent=2)}")
        return DUCDDCStatus

    def XDfeMix_GetMixerStatus(self, device_id):
        """
        Gets Mixer overflow status.

        :param device_id: id of the opened device.
        :return: DUCDDCStatus: DUC/DDC status container.
        """
        self.logger.debug(f"XDfeMix_GetMixerStatus({device_id})")
        MixerStatus = self.MIX.XDfeMix_GetMixerStatus(device_id)
        self.logger.debug(f"MixerStatus = {json.dumps(MixerStatus, indent=2)}")
        return MixerStatus

    def XDfeMix_GetInterruptMask(self, device_id):
        """
        Gets interrupt mask status.

        :param device_id: id of the opened device.
        :return: Mask: interrupt masks container.
        """
        self.logger.debug(f"XDfeMix_GetInterruptMask({device_id})")
        Mask = self.MIX.XDfeMix_GetInterruptMask(device_id)
        self.logger.debug(f"Mask = {json.dumps(Mask, indent=2)}")
        return Mask

    def XDfeMix_SetInterruptMask(self, device_id, Mask):
        """
        Sets interrupt mask.

        :param device_id: id of the opened device.
        :param Mask: interrupt masks flags container.
                     0 - does not mask coresponding interrupt
                     1 - masks coresponding interrupt
        :return: None
        """
        self.logger.debug(f"XDfeMix_SetInterruptMask({device_id}, {json.dumps(Mask, indent=2)})")
        self.MIX.XDfeMix_SetInterruptMask(device_id, Mask)
        return

    def XDfeMix_GetEventStatus(self, device_id):
        """
        Gets event status.

        :param device_id: id of the opened device.
        :return: Status: event status container.
        """
        self.logger.debug(f"XDfeMix_GetEventStatus({device_id})")
        Status = self.MIX.XDfeMix_GetEventStatus(device_id)
        self.logger.debug(f"Status = {json.dumps(Status, indent=2)}")
        return Status

    def XDfeMix_ClearEventStatus(self, device_id, Status):
        """
        Clears event status.

        :param device_id: id of the opened device.
        :param Status: Clear event status container.
                       0 - does not clear coresponding event status
                       1 - clear coresponding event status
        :return: None
        """
        self.logger.debug(f"XDfeMix_ClearEventStatus({device_id}, {json.dumps(Status, indent=2)})")
        self.MIX.XDfeMix_ClearEventStatus(device_id, Status)
        return

    def XDfeMix_SetTUserDelay(self, device_id, Delay):
        """
        Sets the delay, which will be added to TUSER and TLAST (delay matched
        through the IP).

        :param device_id: id of the opened device.
        :param Delay: requested delay variable.
        :return: None
        """
        self.logger.debug(f"XDfeMix_SetTUserDelay({device_id}, {Delay})")
        self.MIX.XDfeMix_SetTUserDelay(device_id, Delay)
        return

    def XDfeMix_GetTUserDelay(self, device_id):
        """
        Reads the delay, which will be added to TUSER and TLAST (delay matched
        through the IP).

        :param device_id: id of the opened device.
        :return: ret: Delay value
        """
        self.logger.debug(f"XDfeMix_GetTUserDelay({device_id})")
        ret = self.MIX.XDfeMix_GetTUserDelay(device_id)
        self.logger.debug(f"ret = {ret}")
        return ret

    def XDfeMix_GetTDataDelay(self, device_id, Tap):
        """
        Returns data latency + tap.

        :param device_id: id of the opened device.
        :param Tap: Tap value.
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
                 TDataDelay: Returned Data latency value.
        """
        self.logger.debug(f"XDfeMix_GetTDataDelay({device_id}, {Tap})")
        ret, TDataDelay = self.MIX.XDfeMix_GetTDataDelay(device_id, Tap)
        self.logger.debug(f"TDataDelay = {json.dumps(TDataDelay, indent=2)}")
        self.logger.debug(f"ret = {ret}")
        return ret, TDataDelay

    def XDfeMix_GetCenterTap(self, device_id, Rate):
        """
        Returns predefined Central Tap value for chosen RATE.

        :param device_id: id of the opened device.
        :param Rate: Interpolation/decimation rate index value [1-5].
        :return: ret: XST_SUCCESS if successful, XST_FAILURE if error occurs.
                 CenterTap: Returned Central Tap value.
        """
        self.logger.debug(f"XDfeMix_GetCenterTap({device_id}, {Rate})")
        ret, CenterTap = self.MIX.XDfeMix_GetCenterTap(device_id, Rate)
        self.logger.debug(f"CenterTap = {json.dumps(CenterTap, indent=2)}")
        self.logger.debug(f"ret = {ret}")
        return ret, CenterTap

    def XDfeMix_SetRegBank(self, device_id, RegBank):
        """
        Sets uplink/downlink register bank.

        :param device_id: id of the opened device.
        :param RegBank: Register bank value to be set.
        :return: None
        """
        self.logger.debug(f"XDfeMix_SetRegBank({device_id}, {RegBank})")
        self.MIX.XDfeMix_SetRegBank(device_id, RegBank)
        return

    def XDfeMix_GetVersions(self, device_id):
        """
        This API gets the driver and HW design version.

        :param device_id: id of the opened device.
        :return: SwVersion: driver version numbers.
                 HwVersion: HW version numbers.
        """
        self.logger.debug(f"XDfeMix_GetVersions({device_id})")
        SwVersion, HwVersion = self.MIX.XDfeMix_GetVersions(device_id)
        self.logger.debug(f"SwVersion = {json.dumps(SwVersion, indent=2)}")
        self.logger.debug(f"HwVersion = {json.dumps(HwVersion, indent=2)}")
        return SwVersion, HwVersion

mix = MIX_Client()
